"""
Module for handling Moliere Potential.

Potential
*********

The Moliere Potential is defined as

.. math::
    U(r) = \\frac{q_i q_j}{4 \\pi \\epsilon_0} \\frac{1}{r} \\sum_{\\alpha} C_{\\alpha} e^{- b_{\\alpha} r}.

For more details see :cite:`Wilson1977`. Note that the parameters :math:`b` are not normalized by the Bohr radius.
They should be passed with the correct units [m] if mks or [cm] if cgs.

Force Error
***********

The force error is calculated from the Yukawa's formula with the smallest screening length.

.. math::

    \\Delta F = \\frac{q^2}{4 \\pi \\epsilon_0} \\sqrt{2 \\pi n b_{\\textrm min} }e^{-b_{\\textrm min} r_c},

This overestimates it, but it doesn't matter.

Potential Attributes
********************
The elements of the :attr:`sarkas.potentials.core.Potential.pot_matrix` are:

.. code-block:: python

    pot_matrix[0] = q_iq_je^2/(4 pi eps_0) Force factor between two particles.
    pot_matrix[1] = C_1
    pot_matrix[2] = C_2
    pot_matrix[3] = C_3
    pot_matrix[4] = b_1
    pot_matrix[5] = b_2
    pot_matrix[6] = b_3

"""
import numpy as np
import numba as nb
from sarkas.utilities.maths import force_error_analytic_pp


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params : sarkas.core.Parameters
        Simulation's parameters.

    """
    potential.screening_lengths = np.array(potential.screening_lengths)
    potential.screening_charges = np.array(potential.screening_charges)
    params_len = len(potential.screening_lengths)

    potential.matrix = np.zeros((2 * params_len + 1, params.num_species, params.num_species))

    for i, q1 in enumerate(params.species_charges):
        for j, q2 in enumerate(params.species_charges):

            potential.matrix[0, i, j] = q1 * q2 / params.fourpie0
            potential.matrix[1 : params_len + 1, i, j] = potential.screening_charges
            potential.matrix[params_len + 1 :, i, j] = potential.screening_lengths

    potential.force = moliere_force
    # Use Yukawa force error formula with the smallest screening length.
    # This overestimates the Force error, but it doesn't matter.
    # The rescaling constant is sqrt ( na^4 ) = sqrt( 3 a/(4pi) )
    params.force_error = force_error_analytic_pp(
        potential.type, potential.rc, potential.matrix[params_len + 1 :, :, :], np.sqrt(3.0 * params.a_ws / (4.0 * np.pi))
    )


@nb.njit
def moliere_force(r, pot_matrix):
    """
    Calculates the PP force between particles using the Moliere Potential.

    Parameters
    ----------
    r : float
        Particles' distance.

    pot_matrix : numpy.ndarray
        Moliere potential parameters. \n
        Shape = (7, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    U : float
        Potential.

    force : float
        Force between two particles.

    """

    U = 0.0
    force = 0.0

    for i in range(int(len(pot_matrix[:-1]) / 2)):
        factor1 = r * pot_matrix[i + 4]
        factor2 = pot_matrix[i + 1] / r
        U += factor2 * np.exp(-factor1)
        force += np.exp(-factor1) * factor2 * (1.0 / r + pot_matrix[i + 1])

    force *= pot_matrix[0]
    U *= pot_matrix[0]

    return U, force



"""
Module for handling Yukawa potential.

Potential
*********

The Yukawa potential between two charges :math:`q_i` and :math:`q_j` at distant :math:`r` is defined as

.. math::
    U_{ab}(r) = \\frac{q_a q_b}{4 \\pi \\epsilon_0} \\frac{e^{- \\kappa r} }{r}.

where :math:`\\kappa = 1/\\lambda` is the screening parameter.

Potential Attributes
********************

The elements of the :attr:`sarkas.potentials.core.Potential.pot_matrix` are:

.. code-block:: python

    pot_matrix[0] = q_iq_j^2/(4 pi eps0)
    pot_matrix[1] = 1/lambda
    pot_matrix[2] = Ewald screening parameter

"""
import numpy as np
from numba import njit
import math as mt
from sarkas.utilities.maths import force_error_analytic_pp


@njit
def yukawa_force_pppm(r, pot_matrix):
    """
    Calculates Potential and Force between two particles when the pppm algorithm is chosen.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : numpy.ndarray
        It contains potential dependent variables. \n
        Shape = (3, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    U : float
        Potential value

    fr : float
        Force between two particles calculated using eq.(22) in :cite:`Dharuman2017`.

    """
    kappa = pot_matrix[1]
    alpha = pot_matrix[2]  # Ewald parameter alpha

    kappa_alpha = kappa / alpha
    alpha_r = alpha * r
    kappa_r = kappa * r
    U = (
        pot_matrix[0]
        * (0.5 / r)
        * (
            np.exp(kappa_r) * mt.erfc(alpha_r + 0.5 * kappa_alpha)
            + np.exp(-kappa_r) * mt.erfc(alpha_r - 0.5 * kappa_alpha)
        )
    )
    # Derivative of the exponential term and 1/r
    f1 = (0.5 / r) * np.exp(kappa * r) * mt.erfc(alpha_r + 0.5 * kappa_alpha) * (1.0 / r - kappa)
    f2 = (0.5 / r) * np.exp(-kappa * r) * mt.erfc(alpha_r - 0.5 * kappa_alpha) * (1.0 / r + kappa)
    # Derivative of erfc(a r) = 2a/sqrt(pi) e^{-a^2 r^2}* (x/r)
    f3 = (alpha / np.sqrt(np.pi) / r) * (
        np.exp(-((alpha_r + 0.5 * kappa_alpha) ** 2)) * np.exp(kappa_r)
        + np.exp(-((alpha_r - 0.5 * kappa_alpha) ** 2)) * np.exp(-kappa_r)
    )
    fr = pot_matrix[0] * (f1 + f2 + f3)

    return U, fr


@njit
def yukawa_force(r, pot_matrix):
    """
    Calculates Potential and Force between two particles.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : numpy.ndarray
        It contains potential dependent variables. \n
        Shape = (3, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    U : float
        Potential.

    force : float
        Force between two particles.

    """
    U = pot_matrix[0] * np.exp(-pot_matrix[1] * r) / r
    force = U * (1.0 / r + pot_matrix[1])

    return U, force


@njit
def force_deriv(r, pot_matrix):
    """Calculate the second derivative of the potential.

    Parameters
    ----------

    r : float
        Distance between particles

    pot_matrix : numpy.ndarray
        Values of the potential constants. \n
        Shape = (3, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    f_dev : float
        Second derivative of potential.

    """
    kappa_r = pot_matrix[1] * r
    U2 = pot_matrix[0] * np.exp(-kappa_r) / r ** 3
    f_dev = U2 * (2.0 * (1.0 + kappa_r) + kappa_r ** 2)
    return f_dev


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params: sarkas.core.Parameters
        Simulation's parameters.

    """

    if potential.method == "pppm":
        potential.matrix = np.zeros((3, params.num_species, params.num_species))
    else:
        potential.matrix = np.zeros((2, params.num_species, params.num_species))

    potential.matrix[1, :, :] = 1.0 / params.lambda_TF

    for i, q1 in enumerate(params.species_charges):
        for j, q2 in enumerate(params.species_charges):
            potential.matrix[0, i, j] = q1 * q2 / params.fourpie0

    if potential.method == "pp":
        # The rescaling constant is sqrt ( na^4 ) = sqrt( 3 a/(4pi) )
        potential.force = yukawa_force
        params.force_error = force_error_analytic_pp(
            potential.type, potential.rc, potential.matrix, np.sqrt(3.0 * params.a_ws / (4.0 * np.pi))
        )
        # # Force error calculated from eq.(43) in Ref.[1]_
        # params.force_error = np.sqrt( TWOPI / params.lambda_TF) * np.exp(- potential.rc / params.lambda_TF)
        # # Renormalize
        # params.force_error *= params.a_ws ** 2 * np.sqrt(params.total_num_ptcls / params.pbox_volume)
    elif potential.method == "pppm":
        potential.force = yukawa_force_pppm
        potential.matrix[2, :, :] = potential.pppm_alpha_ewald
        # PP force error calculation. Note that the equation was derived for a single component plasma.
        kappa_over_alpha = -0.25 * (potential.matrix[1, 0, 0] / potential.matrix[2, 0, 0]) ** 2
        alpha_times_rcut = -((potential.matrix[2, 0, 0] * potential.rc) ** 2)
        params.pppm_pp_err = 2.0 * np.exp(kappa_over_alpha + alpha_times_rcut) / np.sqrt(potential.rc)
        params.pppm_pp_err *= np.sqrt(params.total_num_ptcls) * params.a_ws ** 2 / np.sqrt(params.pbox_volume)



"""
Transport Module.
"""
from IPython import get_ipython

if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# Sarkas Modules
import sarkas.tools.observables as obs
from sarkas.utilities.maths import fast_integral_loop


class TransportCoefficients:
    """
    Transport Coefficients class.
    """

    def __init__(self, params, phase: str = "production", no_slices: int = 1):
        self.__name__ = "TC"
        self.__long_name__ = "Transport Coefficients"
        self.tc_names = ["ElectricalConductivity", "Diffusion", "Interdiffusion", "Viscosities"]
        # Parameters copies
        self.postprocessing_dir = params.postprocessing_dir
        self.units = params.units
        self.job_id = params.job_id
        self.verbose = params.verbose
        self.dt = params.dt
        self.total_plasma_frequency = params.total_plasma_frequency
        self.box_volume = params.box_volume
        self.pbox_volume = params.pbox_volume
        #
        self.saving_dir = None
        self.phase = phase
        self.no_slices = no_slices
        # Calculate the average temperature from the csv data
        self.kB = params.kB
        energy = obs.Thermodynamics()
        energy.setup(params, self.phase)
        energy.parse()
        self.T_avg = energy.dataframe["Temperature"].mean()
        self.beta = 1.0 / (self.kB * self.T_avg)

        self.diffusion_df = None
        self.diffusion_df_slices = None
        self.interdiffusion_df = None
        self.interdiffusion_df_slices = None
        self.viscosity_df = None
        self.viscosity_df_slices = None
        self.conductivity_df = None
        self.conductivity_df_slices = None

        self.create_dir()

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = "Transport( \n"
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ")"
        return disp

    def create_dir(self):
        """Create directory where to save the transport coefficients."""

        saving_dir = os.path.join(self.postprocessing_dir, self.__long_name__.replace(" ", ""))
        if not os.path.exists(saving_dir):
            os.mkdir(saving_dir)

        self.saving_dir = os.path.join(saving_dir, self.phase.capitalize())
        if not os.path.exists(self.saving_dir):
            os.mkdir(self.saving_dir)

    def save_hdf(self, df: pd.DataFrame = None, df_slices: pd.DataFrame = None, tc_name: str = None):
        """
        Save the HDF dataframes to disk in the TransportCoefficient folder.

        Parameters
        ----------
        df: pandas.DataFrame()
            Multi-index dataframe containing the mean and std of the transport coefficient.

        df_slices: pandas.DataFrame()
            Multi-index dataframe containing the transport coefficient of each slice.

        tc_name: str
            Name of the transport coefficient. See :attr:`~.tc_names` for options.

        Returns
        -------
        df: pandas.DataFrame()
            Sorted multi-index dataframe containing the mean and std of the transport coefficient.

        df_slices: pandas.DataFrame()
            Sorted multi-index dataframe containing the transport coefficient of each slice.

        """

        df_slices.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in df_slices.columns])
        df_slices = df_slices.sort_index()
        df_slices.to_hdf(
            os.path.join(self.saving_dir, tc_name + "_slices_" + self.job_id + ".h5"), mode="w", key=tc_name, index=False
        )

        df.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in df.columns])
        df = df.sort_index()
        df.to_hdf(os.path.join(self.saving_dir, tc_name + "_" + self.job_id + ".h5"), mode="w", key=tc_name, index=False)

        return df, df_slices

    def parse(self, observable, tc_name):
        """Read the HDF files containing the transport coefficients.

        Parameters
        ----------
        observable: sarkas.tools.observables.ElectricCurrent
            Observable object containing the ACF whose time integral leads to the electrical conductivity.

        tc_name: str
            Transport Coefficient name.

        Raises
        ------
        ValueError
            Wrong ``tc_name``.

        """
        if tc_name in self.tc_names:

            if tc_name == "ElectricalConductivity":
                self.conductivity_df = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_" + self.job_id + ".h5"), mode="r", index_col=False
                )

                self.conductivity_df_slices = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_slices_" + self.job_id + ".h5"), mode="r", index_col=False
                )

            elif tc_name == "Diffusion":
                self.diffusion_df = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_" + self.job_id + ".h5"), mode="r", index_col=False
                )

                self.diffusion_df_slices = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_slices_" + self.job_id + ".h5"), mode="r", index_col=False
                )

            elif tc_name == "Interdiffusion":

                self.interdiffusion_df = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_" + self.job_id + ".h5"), mode="r", index_col=False
                )

                self.interdiffusion_df_slices = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_slices_" + self.job_id + ".h5"), mode="r", index_col=False
                )

            elif tc_name == "Viscosities":

                self.viscosity_df = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_" + self.job_id + ".h5"), mode="r", index_col=False
                )

                self.viscosity_df_slices = pd.read_hdf(
                    os.path.join(self.saving_dir, tc_name + "_slices_" + self.job_id + ".h5"), mode="r", index_col=False
                )
        else:
            raise ValueError("Wrong tc_name. Please choose amongst ", self.tc_names)

        self.phase = observable.phase
        self.no_slices = observable.no_slices
        self.slice_steps = observable.slice_steps
        self.dump_step = observable.dump_step
        # Print some info
        self.pretty_print(tc_name=tc_name)

    def plot_tc(self, time, acf_data, tc_data, acf_name, tc_name, figname, show: bool = False):
        """
        Make dual plots with ACF and transport coefficient.

        Parameters
        ----------
        acf_data: numpy.ndarray
            Mean and Std of the ACF. \n
            Shape = (:attr:`sarkas.tools.observables.Observable.slice_steps`, 2).

        tc_data: numpy.ndarray
            Mean and Std of the transport coefficient. \n
            Shape = (:attr:`sarkas.tools.observables.Observable.slice_steps`, 2).

        acf_name: str
            y-Label of the ACF plot.

        tc_name: str
            y-label of the transport coefficient plot.

        figname: str
            Name with which to save the plot.

        show: bool
            Flag for displaying the plot if using IPython or terminal.

        Returns
        -------
        fig : matplotlib.figure.Figure
            Figure object.

        (ax1, ax2, ax3, ax4) : tuple
            Tuple with the axes handles. \n
            ax1 = ACF axes, ax2 = transport coefficient axes, ax3 = ax1.twiny(), ax4 = ax2.twiny()

        """
        # Make the plot
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
        ax3 = ax1.twiny()
        ax4 = ax2.twiny()

        # Calculate axis multipliers and labels
        xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, tc_data[:, 0], "Time", tc_name, self.units)

        # ACF
        ax1.plot(xmul * time, acf_data[:, 0] / acf_data[0, 0])
        ax1.fill_between(
            xmul * time,
            (acf_data[:, 0] + acf_data[:, 1]) / (acf_data[0, 0] + acf_data[0, 1]),
            (acf_data[:, 0] - acf_data[:, 1]) / (acf_data[0, 0] - acf_data[0, 1]),
            alpha=0.2,
        )

        # Coefficient
        ax2.plot(xmul * time, ymul * tc_data[:, 0])
        ax2.fill_between(
            xmul * time, ymul * (tc_data[:, 0] + tc_data[:, 1]), ymul * (tc_data[:, 0] - tc_data[:, 1]), alpha=0.2
        )

        xlims = (xmul * time[1], xmul * time[-1] * 1.5)

        ax1.set(xlim=xlims, xscale="log", ylabel=acf_name, xlabel=r"Time difference" + xlbl)
        ax2.set(xlim=xlims, xscale="log", ylabel=tc_name + ylbl, xlabel=r"$\tau$" + xlbl)

        # ax1.legend(loc='best')
        # ax2.legend(loc='best')
        # Finish the index axes
        for axi in [ax3, ax4]:
            axi.grid(alpha=0.1)
            axi.set(xlim=(1, self.slice_steps * 1.5), xscale="log", xlabel="Index")

        fig.tight_layout()
        fig.savefig(os.path.join(self.saving_dir, figname))

        if show:
            fig.show()

        return fig, (ax1, ax2, ax3, ax4)

    def pretty_print(self, tc_name):
        """
        Print to screen the location where data is stored and other relevant information.

        Parameters
        ----------
        observable: sarkas.tools.observables.Observable
            Physical quantity of the ACF.

        tc_name: str
            Name of Transport coefficient to calculate.
        """

        print("Data saved in: \n", os.path.join(self.saving_dir, tc_name + "_" + self.job_id + ".h5"))
        print(os.path.join(self.saving_dir, tc_name + "_slices_" + self.job_id + ".h5"))
        print("\nNo. of slices = {}".format(self.no_slices))
        print("No. dumps per slice = {}".format(int(self.slice_steps / self.dump_step)))

        print(
            "Time interval of autocorrelation function = {:.4e} [s] ~ {} w_p T".format(
                self.dt * self.slice_steps * self.dump_step,
                int(self.dt * self.slice_steps * self.dump_step * self.total_plasma_frequency),
            )
        )

    def electrical_conductivity(self, observable, plot: bool = True, display_plot: bool = False):
        """
        Calculate the transport coefficient from the Green-Kubo formula

        .. math::

            \\sigma = \\frac{\\beta}{V} \\int_0^{\\tau} dt J(t).

        where :math:`\\beta = 1/k_B T` and :math:`V` is the volum of the simulation box.

        Data is retrievable at :attr:`~.conductivity_df` and :attr:`~.conductivity_df_slices`.

        Parameters
        ----------
        observable: sarkas.tools.observables.ElectricCurrent
            Observable object containing the ACF whose time integral leads to the electrical conductivity.

        plot : bool, optional
            Flag for making the dual plot of the ACF and transport coefficient. \n
            Default = True.

        display_plot : bool, optional
            Flag for displaying the plot if using the IPython. Default = False.

        """

        print("\n\n{:=^70} \n".format(" Electrical Conductivity "))
        self.conductivity_df = pd.DataFrame()
        self.conductivity_df_slices = pd.DataFrame()

        # Check that the phase and no_slices is the same from the one computed in the Observable
        observable.parse()

        self.phase = observable.phase
        self.no_slices = observable.no_slices
        self.slice_steps = observable.slice_steps
        self.dump_step = observable.dump_step
        # Print some info
        self.pretty_print(tc_name="ElectricalConductivity")

        # to_numpy creates a 2d-array, hence the [:,0]
        time = observable.dataframe_acf["Time"].iloc[:, 0].to_numpy()

        self.conductivity_df["Time"] = np.copy(time)
        self.conductivity_df_slices["Time"] = np.copy(time)

        jc_str = "Electric Current ACF"
        sigma_str = "Electrical Conductivity"
        const = self.beta / observable.box_volume
        if not observable.magnetized:
            for isl in tqdm(range(observable.no_slices), disable=not observable.verbose):
                integrand = np.array(observable.dataframe_acf_slices[(jc_str, "Total", "slice {}".format(isl))])
                self.conductivity_df_slices[sigma_str + "_slice {}".format(isl)] = const * fast_integral_loop(
                    time, integrand
                )

            col_str = [sigma_str + "_slice {}".format(isl) for isl in range(observable.no_slices)]
            self.conductivity_df[sigma_str + "_Mean"] = self.conductivity_df_slices[col_str].mean(axis=1)
            self.conductivity_df[sigma_str + "_Std"] = self.conductivity_df_slices[col_str].std(axis=1)
        else:

            for isl in tqdm(range(observable.no_slices), disable=not observable.verbose):
                # Parallel
                par_str = (jc_str, "Z", "slice {}".format(isl))
                sigma_par_str = sigma_str + "_Parallel_slice {}".format(isl)
                integrand = observable.dataframe_acf_slices[par_str].to_numpy()
                self.conductivity_df_slices[sigma_par_str] = const * fast_integral_loop(time, integrand)
                # Perpendicular
                x_col_str = (jc_str, "X", "slice {}".format(isl))
                y_col_str = (jc_str, "Y", "slice {}".format(isl))
                perp_integrand = 0.5 * (
                    observable.dataframe_acf_slices[x_col_str].to_numpy()
                    + observable.dataframe_acf_slices[y_col_str].to_numpy()
                )
                sigma_perp_str = sigma_str + "_Perpendicular_slice {}".format(isl)
                self.conductivity_df_slices[sigma_perp_str] = const * fast_integral_loop(time, perp_integrand)

            par_col_str = [(jc_str, "Z", "slice {}".format(isl)) for isl in range(self.no_slices)]
            observable.dataframe_acf[(jc_str, "Parallel", "Mean")] = observable.dataframe_acf_slices[par_col_str].mean(
                axis=1
            )
            observable.dataframe_acf[(jc_str, "Parallel", "Std")] = observable.dataframe_acf_slices[par_col_str].std(
                axis=1
            )

            x_col_str = [(jc_str, "X", "slice {}".format(isl)) for isl in range(self.no_slices)]
            y_col_str = [(jc_str, "Y", "slice {}".format(isl)) for isl in range(self.no_slices)]

            perp_jc = 0.5 * (
                observable.dataframe_acf_slices[x_col_str].to_numpy()
                + observable.dataframe_acf_slices[y_col_str].to_numpy()
            )
            observable.dataframe_acf[(jc_str, "Perpendicular", "Mean")] = perp_jc.mean(axis=1)
            observable.dataframe_acf[(jc_str, "Perpendicular", "Std")] = perp_jc.std(axis=1)
            # Save the updated dataframe
            observable.save_hdf()

            # Average and std of transport coefficient.
            col_str = [sigma_str + "_Parallel_slice {}".format(isl) for isl in range(observable.no_slices)]
            self.conductivity_df[sigma_str + "_Parallel_Mean"] = self.conductivity_df_slices[col_str].mean(axis=1)
            self.conductivity_df[sigma_str + "_Parallel_Std"] = self.conductivity_df_slices[col_str].std(axis=1)
            # Perpendicular
            col_str = [sigma_str + "_Perpendicular_slice {}".format(isl) for isl in range(observable.no_slices)]
            self.conductivity_df[sigma_str + "_Perpendicular_Mean"] = self.conductivity_df_slices[col_str].mean(axis=1)
            self.conductivity_df[sigma_str + "_Perpendicular_Std"] = self.conductivity_df_slices[col_str].std(axis=1)

            # Endif magnetized.

        self.conductivity_df, self.conductivity_df_slices = self.save_hdf(
            df=self.conductivity_df, df_slices=self.conductivity_df_slices, tc_name="ElectricalConductivity"
        )

        if plot or display_plot:

            if not observable.magnetized:
                acf_avg = observable.dataframe_acf[(jc_str, "Total", "Mean")].to_numpy()
                acf_std = observable.dataframe_acf[(jc_str, "Total", "Std")].to_numpy()

                tc_avg = self.conductivity_df[(sigma_str, "Mean")].to_numpy()
                tc_std = self.conductivity_df[(sigma_str, "Std")].to_numpy()

                self.plot_tc(
                    time=time,
                    acf_data=np.column_stack((acf_avg, acf_std)),
                    tc_data=np.column_stack((tc_avg, tc_std)),
                    acf_name="Electric Current ACF",
                    tc_name="Electrical Conductivity",
                    figname="ElectricalConductivity_Plot.png",
                    show=display_plot,
                )
            else:

                acf_avg = observable.dataframe_acf[(jc_str, "Parallel", "Mean")].to_numpy()
                acf_std = observable.dataframe_acf[(jc_str, "Parallel", "Std")].to_numpy()

                tc_avg = self.conductivity_df[(sigma_str, "Parallel", "Mean")].to_numpy()
                tc_std = self.conductivity_df[(sigma_str, "Parallel", "Std")].to_numpy()

                self.plot_tc(
                    time=time,
                    acf_data=np.column_stack((acf_avg, acf_std)),
                    tc_data=np.column_stack((tc_avg, tc_std)),
                    acf_name="Electric Current ACF Parallel",
                    tc_name="Electrical Conductivity Parallel",
                    figname="ElectricalConductivity_Parallel_Plot.png",
                    show=display_plot,
                )

                acf_avg = observable.dataframe_acf[(jc_str, "Perpendicular", "Mean")].to_numpy()
                acf_std = observable.dataframe_acf[(jc_str, "Perpendicular", "Std")].to_numpy()

                tc_avg = self.conductivity_df[(sigma_str, "Perpendicular", "Mean")].to_numpy()
                tc_std = self.conductivity_df[(sigma_str, "Perpendicular", "Std")].to_numpy()

                self.plot_tc(
                    time=time,
                    acf_data=np.column_stack((acf_avg, acf_std)),
                    tc_data=np.column_stack((tc_avg, tc_std)),
                    acf_name="Electric Current ACF Perpendicular",
                    tc_name="Electrical Conductivity Perpendicular",
                    figname="ElectricalConductivity_Perpendicular_Plot.png",
                    show=display_plot,
                )

    def diffusion(
        self, observable, plot: bool = True, display_plot: bool = False,
    ):
        """
        Calculate the transport coefficient from the Green-Kubo formula

        .. math::

            D_{\\alpha} = \\frac{1}{3 N_{\\alpha}} \\sum_{i}^{N_{\\alpha}} \\int_0^{\\tau} dt \\,
                \\langle \\mathbf v^{(\\alpha)}_{i}(t) \\cdot  \\mathbf v^{(\\alpha)}_{i}(0) \\rangle.

        where :math:`\\mathbf v_{i}^{(\\alpha)}(t)` is the velocity of particle :math:`i` of species
        :math:`\\alpha`. Notice that the diffusion coefficient is averaged over all :math:`N_{\\alpha}` particles.

        Data is retrievable at :attr:`~.diffusion_df` and
        :attr:`~.diffusion_df_slices`.

        Parameters
        ----------
        observable: sarkas.tools.observables.ElectricCurrent
            Observable object containing the ACF whose time integral leads to the electrical conductivity.

        plot : bool, optional
            Flag for making the dual plot of the ACF and transport coefficient. Default = True.

        display_plot : bool, optional
            Flag for displaying the plot if using the IPython. Default = False.

        """
        print("\n\n{:=^70} \n".format(" Diffusion Coefficient "))
        self.diffusion_df = pd.DataFrame()
        self.diffusion_df_slices = pd.DataFrame()

        # Check that the phase and no_slices is the same from the one computed in the Observable
        observable.parse()

        self.phase = observable.phase
        self.no_slices = observable.no_slices
        self.slice_steps = observable.slice_steps
        self.dump_step = observable.dump_step
        # Print some info
        self.pretty_print(tc_name="Diffusion")

        # to_numpy creates a 2d-array, hence the [:,0]
        time = observable.dataframe_acf["Time"].iloc[:, 0].to_numpy()

        self.diffusion_df["Time"] = np.copy(time)
        self.diffusion_df_slices["Time"] = np.copy(time)

        vacf_str = "VACF"
        const = 1.0 / 3.0

        if not observable.magnetized:
            # Loop over time slices
            for isl in tqdm(range(self.no_slices), disable=not observable.verbose):

                # Iterate over the number of species
                for i, sp in enumerate(observable.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    # Grab vacf data of each slice
                    integrand = np.array(observable.dataframe_acf_slices[(sp_vacf_str, "Total", "slice {}".format(isl))])
                    df_str = "{} Diffusion_slice {}".format(sp, isl)
                    self.diffusion_df_slices[df_str] = const * fast_integral_loop(time=time, integrand=integrand)

            # Average and std of each diffusion coefficient.
            for isp, sp in enumerate(observable.species_names):
                col_str = ["{} Diffusion_slice {}".format(sp, isl) for isl in range(observable.no_slices)]

                self.diffusion_df["{} Diffusion_Mean".format(sp)] = self.diffusion_df_slices[col_str].mean(axis=1)
                self.diffusion_df["{} Diffusion_Std".format(sp)] = self.diffusion_df_slices[col_str].std(axis=1)

        else:
            # Loop over time slices
            for isl in tqdm(range(observable.no_slices), disable=not observable.verbose):

                # Iterate over the number of species
                for i, sp in enumerate(observable.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str

                    # Parallel
                    par_vacf_str = (sp_vacf_str, "Z", "slice {}".format(isl))
                    integrand_par = observable.dataframe_acf_slices[par_vacf_str].to_numpy()
                    par_slice_str = "{} Diffusion_Parallel_slice {}".format(sp, isl)
                    self.diffusion_df_slices[par_slice_str] = fast_integral_loop(time=time, integrand=integrand_par)
                    # Perpendicular
                    perp_vacf_str = (sp_vacf_str, "X", "slice {}".format(isl))
                    perp_slice_str = "{} Diffusion_Perpendicular_slice {}".format(sp, isl)
                    integrand_perp = (
                        observable.dataframe_acf_slices[perp_vacf_str].to_numpy()
                        + observable.dataframe_acf_slices[perp_vacf_str].to_numpy()
                    )

                    self.diffusion_df_slices[perp_slice_str] = 0.5 * fast_integral_loop(
                        time=time, integrand=integrand_perp
                    )

            # Add the average and std of perp and par VACF to its dataframe
            for isp, sp in enumerate(observable.species_names):
                sp_vacf_str = "{} ".format(sp) + vacf_str
                sp_diff_str = "{} ".format(sp) + "Diffusion"
                par_col_str = [(sp_vacf_str, "Z", "slice {}".format(isl)) for isl in range(self.no_slices)]

                observable.dataframe_acf[(sp_vacf_str, "Parallel", "Mean")] = observable.dataframe_acf_slices[
                    par_col_str
                ].mean(axis=1)
                observable.dataframe_acf[(sp_vacf_str, "Parallel", "Std")] = observable.dataframe_acf_slices[
                    par_col_str
                ].std(axis=1)

                x_col_str = [(sp_vacf_str, "X", "slice {}".format(isl)) for isl in range(self.no_slices)]
                y_col_str = [(sp_vacf_str, "Y", "slice {}".format(isl)) for isl in range(self.no_slices)]

                perp_vacf = 0.5 * (
                    observable.dataframe_acf_slices[x_col_str].to_numpy()
                    + observable.dataframe_acf_slices[y_col_str].to_numpy()
                )
                observable.dataframe_acf[(sp_vacf_str, "Perpendicular", "Mean")] = perp_vacf.mean(axis=1)
                observable.dataframe_acf[(sp_vacf_str, "Perpendicular", "Std")] = perp_vacf.std(axis=1)

                # Average and std of each diffusion coefficient.
                par_col_str = [sp_diff_str + "_Parallel_slice {}".format(isl) for isl in range(self.no_slices)]
                perp_col_str = [sp_diff_str + "_Perpendicular_slice {}".format(isl) for isl in range(self.no_slices)]

                self.diffusion_df[sp_diff_str + "_Parallel_Mean"] = self.diffusion_df_slices[par_col_str].mean(axis=1)
                self.diffusion_df[sp_diff_str + "_Parallel_Std"] = self.diffusion_df_slices[par_col_str].std(axis=1)

                self.diffusion_df[sp_diff_str + "_Perpendicular_Mean"] = self.diffusion_df_slices[perp_col_str].mean(
                    axis=1
                )
                self.diffusion_df[sp_diff_str + "_Perpendicular_Std"] = self.diffusion_df_slices[perp_col_str].std(axis=1)

            # Save the updated dataframe
            observable.save_hdf()
            # Endif magnetized.

        self.diffusion_df, self.diffusion_df_slices = self.save_hdf(
            df=self.diffusion_df, df_slices=self.diffusion_df_slices, tc_name="diffusion"
        )

        if plot or display_plot:

            if observable.magnetized:
                for isp, sp in enumerate(observable.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    sp_diff_str = "{} ".format(sp) + "Diffusion"

                    # Parallel
                    acf_avg = observable.dataframe_acf[(sp_vacf_str, "Parallel", "Mean")].to_numpy()
                    acf_std = observable.dataframe_acf[(sp_vacf_str, "Parallel", "Std")].to_numpy()

                    tc_avg = self.diffusion_df[(sp_diff_str, "Parallel", "Mean")].to_numpy()
                    tc_std = self.diffusion_df[(sp_diff_str, "Parallel", "Std")].to_numpy()

                    self.plot_tc(
                        time=time,
                        acf_data=np.column_stack((acf_avg, acf_std)),
                        tc_data=np.column_stack((tc_avg, tc_std)),
                        acf_name=sp_vacf_str + " Parallel",
                        tc_name=sp_diff_str + " Parallel",
                        figname="{}_Parallel_Diffusion_Plot.png".format(sp),
                        show=display_plot,
                    )

                    # Perpendicular
                    acf_avg = observable.dataframe_acf[(sp_vacf_str, "Perpendicular", "Mean")]
                    acf_std = observable.dataframe_acf[(sp_vacf_str, "Perpendicular", "Std")]

                    tc_avg = self.diffusion_df[(sp_diff_str, "Perpendicular", "Mean")]
                    tc_std = self.diffusion_df[(sp_diff_str, "Perpendicular", "Std")]

                    self.plot_tc(
                        time=time,
                        acf_data=np.column_stack((acf_avg, acf_std)),
                        tc_data=np.column_stack((tc_avg, tc_std)),
                        acf_name=sp_vacf_str + " Perpendicular",
                        tc_name=sp_diff_str + " Perpendicular",
                        figname="{}_Perpendicular_Diffusion_Plot.png".format(sp),
                        show=display_plot,
                    )
            else:
                for isp, sp in enumerate(observable.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    acf_avg = observable.dataframe_acf[(sp_vacf_str, "Total", "Mean")].to_numpy()
                    acf_std = observable.dataframe_acf[(sp_vacf_str, "Total", "Std")].to_numpy()

                    d_str = "{} Diffusion".format(sp)
                    tc_avg = self.diffusion_df[(d_str, "Mean")].to_numpy()
                    tc_std = self.diffusion_df[(d_str, "Std")].to_numpy()

                    self.plot_tc(
                        time=time,
                        acf_data=np.column_stack((acf_avg, acf_std)),
                        tc_data=np.column_stack((tc_avg, tc_std)),
                        acf_name=sp_vacf_str,
                        tc_name=d_str,
                        figname="{}_Diffusion_Plot.png".format(sp),
                        show=display_plot,
                    )

    def interdiffusion(self, observable, plot: bool = True, display_plot: bool = False):
        """
        Calculate the transport coefficient from the Green-Kubo formula

        .. math::

            D_{\\alpha} = \\frac{1}{3Nx_1x_2} \\int_0^\\tau dt
            \\langle \\mathbf {J}_{\\alpha}(0) \\cdot \\mathbf {J}_{\\alpha}(t) \\rangle,

        where :math:`x_{1,2}` are the concentration of the two species and
        :math:`\\mathbf {J}_{\\alpha}(t)` is the diffusion current calculated by the
        :class:`sarkas.tools.observables.DiffusionFlux` class.

        Data is retrievable at :attr:`~.interdiffusion_df` and :attr:`~.interdiffusion_df_slices`.

        Parameters
        ----------
        observable: sarkas.tools.observables.DiffusionFlux
            Observable object containing the ACF whose time integral leads to the interdiffusion coefficient.

        plot : bool, optional
            Flag for making the dual plot of the ACF and transport coefficient. Default = True.

        display_plot : bool, optional
            Flag for displaying the plot if using the IPython. Default = False

        """

        print("\n\n{:=^70} \n".format(" Interdiffusion Coefficient "))
        self.interdiffusion_df = pd.DataFrame()
        self.interdiffusion_df_slices = pd.DataFrame()

        # Check that the phase and no_slices is the same from the one computed in the Observable
        observable.parse()

        self.phase = observable.phase
        self.no_slices = observable.no_slices
        self.slice_steps = observable.slice_steps
        self.dump_step = observable.dump_step
        # Print some info
        self.pretty_print(tc_name="Interdiffusion")

        # to_numpy creates a 2d-array, hence the [:,0]
        time = observable.dataframe_acf["Time"].to_numpy()[:, 0]
        self.interdiffusion_df["Time"] = np.copy(time)
        self.interdiffusion_df_slices["Time"] = np.copy(time)

        no_fluxes_acf = observable.no_fluxes_acf
        # Normalization constant
        const = 1.0 / (3.0 * observable.total_num_ptcls * observable.species_concentrations.prod())

        df_str = "Diffusion Flux ACF"
        id_str = "Inter Diffusion Flux"
        for isl in tqdm(range(self.no_slices), disable=not self.verbose):
            # D_ij = np.zeros((no_fluxes_acf, jc_acf.slice_steps))

            for ij in range(no_fluxes_acf):
                acf_df_str = (df_str + " {}".format(ij), "Total", "slice {}".format(isl))
                integrand = observable.dataframe_acf_slices[acf_df_str].to_numpy()

                id_df_str = id_str + " {}_slice {}".format(ij, isl)
                self.interdiffusion_df_slices[id_df_str] = const * fast_integral_loop(time=time, integrand=integrand)

        # Average and Std of slices
        for ij in range(no_fluxes_acf):
            col_str = [id_str + " {}_slice {}".format(ij, isl) for isl in range(self.no_slices)]
            self.interdiffusion_df[id_str + " {}_Mean".format(ij)] = self.interdiffusion_df_slices[col_str].mean(axis=1)
            self.interdiffusion_df[id_str + " {}_Std".format(ij)] = self.interdiffusion_df_slices[col_str].std(axis=1)

        self.interdiffusion_df, self.interdiffusion_df_slices = self.save_hdf(
            df=self.interdiffusion_df, df_slices=self.interdiffusion_df_slices, tc_name="Interdiffusion"
        )

        if plot or display_plot:

            for flux in range(observable.no_fluxes_acf):
                flux_str = df_str + " {}".format(flux)
                acf_avg = observable.dataframe_acf[(flux_str, "Total", "Mean")].to_numpy()
                acf_std = observable.dataframe_acf[(flux_str, "Total", "Std")].to_numpy()

                d_str = id_str + " {}".format(flux)
                tc_avg = self.interdiffusion_df[(d_str, "Mean")].to_numpy()
                tc_std = self.interdiffusion_df[(d_str, "Std")].to_numpy()

                self.plot_tc(
                    time=time,
                    acf_data=np.column_stack((acf_avg, acf_std)),
                    tc_data=np.column_stack((tc_avg, tc_std)),
                    acf_name=flux_str,
                    tc_name=d_str,
                    figname="InterDiffusion_{}_Plot.png".format(flux),
                    show=display_plot,
                )

    def viscosity(self, observable, plot: bool = True, display_plot: bool = False):
        """
        Calculate the transport coefficient from the Green-Kubo formula.

        The shear viscosity is obtained from

        .. math::

            \\eta = \\frac{\\beta V}{6} \\sum_{\\alpha} \\sum_{\\gamma \\neq \\alpha} \\int_0^{\\tau} dt \\,
            \\left \\langle \\mathcal P_{\\alpha\\gamma}(t) \\mathcal P_{\\alpha\\gamma}(0) \\right \\rangle

        where :math:`\\beta = 1/k_B T`, :math:`\\alpha,\\gamma = {x, y, z}` and
        :math:`\\mathcal P_{\\alpha\\gamma}(t)` is the element of the Pressure Tensor calculated with
        :class:`sarkas.tools.observables.PressureTensor`.

        The bulk viscosity is obtained from

        .. math::

            \\eta_V = \\beta V \\int_0^{\\tau}dt \\,
             \\left \\langle \\delta \\mathcal P(t) \\delta \\mathcal P(0) \\right \\rangle,

        where

        .. math::
            \\delta \\mathcal P(t) = \\mathcal P(t) - \\left \\langle \\mathcal P  \\right \\rangle

        is the deviation of the scalar pressure.

        Parameters
        ----------
        observable: sarkas.tools.observables.DiffusionFlux
            Observable object containing the ACF whose time integral leads to the interdiffusion coefficient.

        plot : bool, optional
            Flag for making the dual plot of the ACF and transport coefficient. Default = True.

        display_plot : bool, optional
            Flag for displaying the plot if using the IPython. Default = False

        """
        print("\n\n{:=^70} \n".format(" Viscosity Coefficient "))
        self.viscosity_df = pd.DataFrame()
        self.viscosity_df_slices = pd.DataFrame()

        # Check that the phase and no_slices is the same from the one computed in the Observable
        observable.parse()

        self.phase = observable.phase
        self.no_slices = observable.no_slices
        self.slice_steps = observable.slice_steps
        self.dump_step = observable.dump_step

        # Print some info
        self.pretty_print(tc_name="Viscosity")

        # to_numpy creates a 2d-array, hence the [:,0]
        time = observable.dataframe_acf["Time"].iloc[:, 0].to_numpy()
        self.viscosity_df["Time"] = np.copy(time)
        self.viscosity_df_slices["Time"] = np.copy(time)

        dim_lbl = ["x", "y", "z"]

        pt_str_list = [
            "Pressure Tensor Kinetic ACF",
            "Pressure Tensor Potential ACF",
            "Pressure Tensor Kin-Pot ACF",
            "Pressure Tensor Pot-Kin ACF",
            "Pressure Tensor ACF",
        ]
        eta_str_list = [
            "Shear Viscosity Tensor Kinetic",
            "Shear Viscosity Tensor Potential",
            "Shear Viscosity Tensor Kin-Pot",
            "Shear Viscosity Tensor Pot-Kin",
            "Shear Viscosity Tensor Total",
        ]

        start_steps = 0
        end_steps = 0

        for isl in tqdm(range(self.no_slices), disable=not observable.verbose):
            end_steps += observable.slice_steps

            const = observable.box_volume * self.beta
            # Calculate Bulk Viscosity
            # It is is calculated from the fluctuations of the pressure eq. 2.124a Allen & Tilsdeley
            integrand = observable.dataframe_acf_slices[("Delta Pressure ACF", "slice {}".format(isl))].to_numpy()
            self.viscosity_df_slices["Bulk Viscosity_slice {}".format(isl)] = const * fast_integral_loop(time, integrand)

            # Calculate the Shear Viscosity Elements
            for _, ax1 in enumerate(dim_lbl):
                for _, ax2 in enumerate(dim_lbl):
                    for _, (pt_str, eta_str) in enumerate(zip(pt_str_list, eta_str_list)):
                        pt_str_temp = (pt_str + " {}{}{}{}".format(ax1, ax2, ax1, ax2), "slice {}".format(isl))
                        integrand = observable.dataframe_acf_slices[pt_str_temp].to_numpy()
                        eta_str_temp = eta_str + " {}{}_slice {}".format(ax1, ax2, isl)
                        self.viscosity_df_slices[eta_str_temp] = const * fast_integral_loop(time, integrand)

            start_steps += observable.slice_steps

        # Now average the slices
        col_str = ["Bulk Viscosity_slice {}".format(isl) for isl in range(observable.no_slices)]
        self.viscosity_df["Bulk Viscosity_Mean"] = self.viscosity_df_slices[col_str].mean(axis=1)
        self.viscosity_df["Bulk Viscosity_Std"] = self.viscosity_df_slices[col_str].std(axis=1)

        for _, ax1 in enumerate(dim_lbl):
            for _, ax2 in enumerate(dim_lbl):
                for _, eta_str in enumerate(eta_str_list):
                    col_str = [eta_str + " {}{}_slice {}".format(ax1, ax2, isl) for isl in range(observable.no_slices)]
                    self.viscosity_df[eta_str + " {}{}_Mean".format(ax1, ax2)] = self.viscosity_df_slices[col_str].mean(
                        axis=1
                    )
                    self.viscosity_df[eta_str + " {}{}_Std".format(ax1, ax2)] = self.viscosity_df_slices[col_str].std(
                        axis=1
                    )

        list_coord = ["xy", "xz", "yx", "yz", "zx", "zy"]
        col_str = [eta_str + " {}_Mean".format(coord) for coord in list_coord]
        self.viscosity_df["Shear Viscosity_Mean"] = self.viscosity_df[col_str].mean(axis=1)
        self.viscosity_df["Shear Viscosity_Std"] = self.viscosity_df[col_str].std(axis=1)

        self.viscosity_df, self.viscosity_df_slices = self.save_hdf(
            df=self.viscosity_df, df_slices=self.viscosity_df_slices, tc_name="Viscosities"
        )

        plot_quantities: list = ["Bulk Viscosity", "Shear Viscosity"]
        if plot or display_plot:
            # Make the plot
            for ipq, pq in enumerate(plot_quantities):
                if pq == "Bulk Viscosity":
                    acf_str = "Delta Pressure ACF"
                    acf_avg = observable.dataframe_acf[("Delta Pressure ACF", "Mean")]
                    acf_std = observable.dataframe_acf[("Delta Pressure ACF", "Std")]
                elif pq == "Shear Viscosity":
                    # The axis are the last two elements in the string
                    acf_strs = [("Pressure Tensor ACF {}".format(coord), "Mean") for coord in list_coord]
                    acf_avg = observable.dataframe_acf[acf_strs].mean(axis=1)
                    acf_std = observable.dataframe_acf[acf_strs].std(axis=1)

                tc_avg = self.viscosity_df[(pq, "Mean")]
                tc_std = self.viscosity_df[(pq, "Std")]

                self.plot_tc(
                    time=time,
                    acf_data=np.column_stack((acf_avg, acf_std)),
                    tc_data=np.column_stack((tc_avg, tc_std)),
                    acf_name=acf_str,
                    tc_name=pq,
                    figname="{}_Plot.png".format(pq),
                    show=display_plot,
                )



"""
Module for handling Quantum Statistical Potentials.

Potential
*********

Quantum Statistical Potentials are defined by three terms

.. math::
    U(r) = U_{\\rm pauli}(r) + U_{\\rm coul} + U_{\\rm diff} (r)

where

.. math::
    U_{\\rm pauli}(r) = k_BT \\ln (2)  e^{ - 4\\pi r^2/ \\Lambda_{ab}^2 }

is due to the Pauli exclusion principle,

.. math::
    U_{\\rm coul}(r) = \\frac{q_iq_j}{4\\pi \\epsilon_0} \\frac{1}{r}

is the usual Coulomb interaction, and :math:`U_{\\rm diff}(r)` is a diffraction term.

There are two possibilities for the diffraction term. The most common is the Deutsch Potential

.. math::
    U_{\\rm deutsch}(r) = \\frac{q_aq_b}{4\\pi \\epsilon_0} \\frac{e^{- 2 \\pi r/\\Lambda_{ab}} }{r}.

The second most common form is the Kelbg potential

.. math::
    U_{\\rm kelbg}(r) = - \\frac{q_aq_b}{4\\pi \\epsilon_0} \\frac{1}{r} \\left [  e^{- 2 \\pi r^2/\\Lambda_{ab}^2 }
    - \\sqrt{2} \\pi \\dfrac{r}{\\Lambda_{ab}} \\textrm{erfc} \\left ( \\sqrt{ 2\\pi}  r/ \\Lambda_{ab} \\right )
    \\right ].

In the above equations the screening length :math:`\\Lambda_{ab}` is the thermal de Broglie wavelength
between the two charges defined as

.. math::
   \\Lambda_{ab} = \\sqrt{\\frac{2\\pi \\hbar^2}{\\mu_{ab} k_BT}}, \\quad  \\mu_{ab} = \\frac{m_a m_b}{m_a + m_b}


Note that in Ref. :cite:`Hansen1981` the DeBroglie wavelength is defined as

.. math::
   \\Lambda_{ee} = \\sqrt{ \\dfrac{\\hbar^2}{2 \\pi \\mu_{ee} k_{B} T}},

while in statistical physics textbooks is defined as

.. math::
   \\Lambda_{ee} = \\sqrt{ \\dfrac{2 \\pi \\hbar^2}{\\mu_{ee} k_{B} T}} .

The latter will be used in Sarkas. The difference is in the factor of :math:`2\\pi`, i.e. the difference between
a wave number and wave length.

Potential Attributes
********************

The elements of the :attr:`sarkas.potentials.core.Potential.pot_matrix` are:

.. code-block:: python

    pot_matrix[0] = qi*qj/4*pi*eps0
    pot_matrix[1] = 2pi/deBroglie
    pot_matrix[2] = e-e Pauli term factor
    pot_matrix[3] = e-e Pauli term exponent term
    pot_matrix[4] = Ewald parameter
    pot_matrix[5] = Short-range cutoff

"""

from warnings import warn
import numpy as np
from numba import njit
import math as mt

from sarkas.utilities.exceptions import AlgorithmWarning
from sarkas.utilities.maths import force_error_analytic_pp, TWOPI


def update_params(potential, params):
    """
    Create potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params : object
        Simulation's parameters


    """
    # Do a bunch of checks
    # pppm algorithm only
    if potential.method != "pppm":
        raise ValueError("QSP interaction can only be calculated using pppm algorithm.")

    # Check for neutrality
    if params.total_net_charge != 0:
        warn("Total net charge is not zero.", category=AlgorithmWarning)

    # Default attributes
    if not hasattr(potential, "qsp_type"):
        potential.qsp_type = "deutsch"
    if not hasattr(potential, "qsp_pauli"):
        potential.qsp_pauli = True

    # Enforce consistency
    potential.qsp_type = potential.qsp_type.lower()

    four_pi = 2.0 * TWOPI
    log2 = np.log(2.0)

    # Redefine ion temperatures and ion total number density
    mask = params.species_names != "e"
    params.total_ion_temperature = params.species_concentrations[mask].transpose() * params.species_temperatures[mask]
    params.ni = np.sum(params.species_num_dens[mask])

    # Calculate the total and ion Wigner-Seitz Radius from the total density
    params.ai = (3.0 / (four_pi * params.ni)) ** (1.0 / 3.0)  # Ion WS

    deBroglie_const = TWOPI * params.hbar2 / params.kB

    potential.matrix = np.zeros((6, params.num_species, params.num_species))
    for i, name1 in enumerate(params.species_names):
        m1 = params.species_masses[i]
        q1 = params.species_charges[i]

        for j, name2 in enumerate(params.species_names):
            m2 = params.species_masses[j]
            q2 = params.species_charges[j]

            reduced = (m1 * m2) / (m1 + m2)

            if name1 == "e" or name2 == "e":
                # Use electron temperature in e-e and e-i interactions
                lambda_deB = np.sqrt(deBroglie_const / (reduced * params.electron_temperature))
            else:
                # Use ion temperature in i-i interactions only
                lambda_deB = np.sqrt(deBroglie_const / (reduced * params.total_ion_temperature))

            if name2 == name1:  # e-e
                potential.matrix[2, i, j] = log2 * params.kB * params.electron_temperature
                potential.matrix[3, i, j] = four_pi / (log2 * lambda_deB ** 2)

            potential.matrix[0, i, j] = q1 * q2 / params.fourpie0
            potential.matrix[1, i, j] = TWOPI / lambda_deB

    if not potential.qsp_pauli:
        potential.matrix[2, :, :] = 0.0

    potential.matrix[4, :, :] = potential.pppm_alpha_ewald
    potential.matrix[5, :, :] = potential.a_rs

    if potential.qsp_type == "deutsch":
        potential.force = deutsch_force
        # Calculate the PP Force error from the e-e diffraction term only.
        params.pppm_pp_err = force_error_analytic_pp(
            potential.type, potential.rc, potential.matrix, np.sqrt(3.0 * params.a_ws / (4.0 * np.pi))
        )
    elif potential.qsp_type == "kelbg":
        potential.force = kelbg_force
        # TODO: Calculate the PP Force error from the e-e diffraction term only.
        # the following is a placeholder
        params.pppm_pp_err = force_error_analytic_pp(
            potential.type, potential.rc, potential.matix, np.sqrt(3.0 * params.a_ws / (4.0 * np.pi))
        )


@njit
def deutsch_force(r_in, pot_matrix):
    """
    Calculate Deutsch QSP Force between two particles.

    Parameters
    ----------
    r_in : float
        Distance between two particles.

    pot_matrix : numpy.ndarray
        It contains potential dependent variables. \n
        Shape = (6, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    U : float
        Potential.

    force : float
        Force between two particles.


    """

    A = pot_matrix[0]
    C = pot_matrix[1]
    D = pot_matrix[2]
    F = pot_matrix[3]
    alpha = pot_matrix[4]
    rs = pot_matrix[5]

    # Branchless programming
    r = r_in * (r_in >= rs) + rs * (r_in < rs)

    a2 = alpha * alpha
    r2 = r * r

    # Ewald short-range potential and force terms
    U_ewald = A * mt.erfc(alpha * r) / r
    f_ewald = U_ewald / r  # 1/r derivative
    f_ewald += A * (2.0 * alpha / np.sqrt(np.pi)) * np.exp(-a2 * r2) / r  # erfc derivative

    # Diffraction potential and force term
    U_diff = -A * np.exp(-C * r) / r
    f_diff = U_diff / r  # 1/r derivative
    f_diff += -A * C * np.exp(-C * r) / r  # exp derivative

    # Pauli potential and force terms
    U_pauli = D * np.exp(-F * r2)
    f_pauli = 2.0 * r * D * F * np.exp(-F * r2)

    U = U_ewald + U_diff + U_pauli
    force = f_ewald + f_diff + f_pauli

    return U, force


@njit
def kelbg_force(r_in, pot_matrix):
    """
    Calculates the QSP Force between two particles when the pppm algorithm is chosen.

    Parameters
    ----------
    r_in : float
        Distance between two particles.

    pot_matrix : numpy.ndarray
        It contains potential dependent variables. \n
        Shape = (6, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    U : float
        Potential.

    force : float
        Force between two particles.

    """

    A = pot_matrix[0]
    C = pot_matrix[1]
    D = pot_matrix[2]
    F = pot_matrix[3]
    alpha = pot_matrix[4]
    rs = pot_matrix[5]

    # Branchless programming
    r = r_in * (r_in >= rs) + rs * (r_in < rs)

    C2 = C * C
    a2 = alpha * alpha
    r2 = r * r

    # Ewald short-range potential and force terms
    U_ewald = A * mt.erfc(alpha * r) / r
    f_ewald = U_ewald / r2  # 1/r derivative
    f_ewald += A * (2.0 * alpha / np.sqrt(np.pi) / r2) * np.exp(-a2 * r2)  # erfc derivative

    # Diffraction potential and force term
    U_diff = -A * np.exp(-C2 * r2 / np.pi) / r
    U_diff += A * C * mt.erfc(C * r / np.sqrt(np.pi))

    f_diff = -A * (2.0 * C2 * r2 + np.pi) * np.exp(-C2 * r2 / np.pi) / (np.pi * r * r2)  # exp(r)/r derivative
    f_diff += 2.0 * A * C2 * np.exp(-C2 * r2 / np.pi) / r / np.pi  # erfc derivative

    # Pauli Term
    U_pauli = D * np.exp(-F * r2)
    f_pauli = 2.0 * D * F * np.exp(-F * r2)

    U = U_ewald + U_diff + U_pauli
    force = f_ewald + f_diff + f_pauli

    return U, force



"""
Module for handling Exact Gradient corrected Screened (EGS) Potential.

Potential
*********

The exact-gradient screened (EGS) potential introduces new parameters that can be easily calculated from initial inputs.
Density gradient corrections to the free energy functional lead to the first parameter, :math:`\nu`,

.. math::
   \\nu = - \\frac{3\\lambda}{\\pi^{3/2}}  \\frac{4\\pi \\bar{e}^2 \\beta }{\\Lambda_{e}} \\frac{d}{d\\eta} \\mathcal I_{-1/2}(\\eta),

where :math:`\\lambda` is a correction factor; :math:`\\lambda = 1/9` for the true gradient corrected Thomas-Fermi model
and :math:`\\lambda = 1` for the traditional von Weissaecker model, :math:`\\mathcal I_{-1/2}[\\eta_0]` is the
:ref:`Fermi Integral` of order :math:`-1/2`, and :math:`\\Lambda_e` is the :ref:`de Broglie wavelength` of the electrons.

In the case :math:`\\nu < 1` the EGS potential takes the form

.. math::
   U_{ab}(r) = \\frac{Z_a Z_b \\bar{e}^2 }{2r}\\left [ ( 1+ \\alpha ) e^{-r/\\lambda_-} + ( 1 - \\alpha) e^{-r/\\lambda_+} \\right ],

with

.. math::
   \\lambda_\\pm^2 = \\frac{\\nu \\lambda_{\\textrm{TF}}^2}{2b \\pm 2b\\sqrt{1 - \\nu}}, \\quad \\alpha = \\frac{b}{\\sqrt{b - \\nu}},

where the parameter :math:`b` arises from exchange-correlation contributions, see below.\n
On the other hand :math:`\\nu > 1`, the pair potential has the form

.. math::
   U_{ab}(r) = \\frac{Z_a Z_b \\bar{e}^2}{r}\\left [ \\cos(r/\\gamma_-) + \\alpha' \\sin(r/\\gamma_-) \\right ] e^{-r/\\gamma_+}

with

.. math::
   \\gamma_\\pm^2 = \\frac{\\nu\\lambda_{\\textrm{TF}}^2}{\\sqrt{\\nu} \\pm b}, \\quad \\alpha' = \\frac{b}{\\sqrt{\\nu - b}}.

Neglect of exchange-correlational effects leads to :math:`b = 1` otherwise

.. math::
   b = 1 - \\frac{2}{8} \\frac{1}{k_{\\textrm{F}}^2 \\lambda_{\\textrm{TF}}^2 }  \\left [ h\\left ( \\Theta \\right ) - 2 \\Theta h'(\\Theta) \\right ]

where :math:`k_{\\textrm{F}}` is the Fermi wavenumber and :math:`\\Theta = (\\beta E_{\\textrm{F}})^{-1}` is the electron
:ref:`Degeneracy Parameter` calculated from the :ref:`Fermi Energy`.

.. math::
   h \\left ( \\Theta \\right) = \\frac{N(\\Theta)}{D(\\Theta)}\\tanh \\left( \\Theta^{-1} \\right ),

.. math::
   N(\\Theta) = 1 + 2.8343\\Theta^2 - 0.2151\\Theta^3 + 5.2759\\Theta^4,

.. math::
   D \\left ( \\Theta \\right ) = 1 + 3.9431\\Theta^2 + 7.9138\\Theta^4.

Force Error
***********

The EGS potential is always smaller than pure Yukawa. Therefore the force error is chosen to be the same as Yukawa's

.. math::

    \\Delta F = \\frac{q^2}{4 \\pi \\epsilon_0} \\sqrt{\\frac{2 \\pi n}{\\lambda_{-}}}e^{-r_c/\\lambda_-}

This overestimates it, but it doesn't matter.

Potential Attributes
********************

The elements of :attr:`sarkas.potentials.core.Potential.pot_matrix` are

if :attr:`sarkas.core.Parameters.nu` less than 1:

.. code-block::

    pot_matrix[0] = q_iq_j/4pi eps0
    pot_matrix[1] = nu
    pot_matrix[2] = 1 + alpha
    pot_matrix[3] = 1 - alpha
    pot_matrix[4] = 1.0 / lambda_minus
    pot_matrix[5] = 1.0 / lambda_plus

else

.. code-block::

    pot_matrix[0] = q_iq_j/4pi eps0
    pot_matrix[1] = nu
    pot_matrix[2] = 1.0
    pot_matrix[3] = alpha prime
    pot_matrix[4] = 1.0 / gamma_minus
    pot_matrix[5] = 1.0 / gamma_plus

"""
import numpy as np
from numba import njit
import fdint

from sarkas.utilities.exceptions import AlgorithmError
from sarkas.utilities.maths import force_error_analytic_pp


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params: sarkas.core.Parameters
        Simulation's parameters.

    Raises
    ------
    `~sarkas.utilities.exceptions.AlgorithmError`
        If the chosen algorithm is pppm.

    """

    # lambda factor : 1 = von Weizsaecker, 1/9 = Thomas-Fermi
    if not hasattr(potential, "lmbda"):
        potential.lmbda = 1.0 / 9.0

    fdint_dfdk_vec = np.vectorize(fdint.dfdk)
    # eq. (14) of Ref. [1]_
    params.nu = 3.0 / np.pi ** 1.5 * params.landau_length / params.lambda_deB
    params.nu *= potential.lmbda * fdint_dfdk_vec(k=-0.5, phi=params.eta_e)

    # Degeneracy Parameter
    theta = params.electron_degeneracy_parameter
    if 0.1 <= theta <= 12:
        # Regime of validity of the following approximation Perrot et al. Phys Rev A 302619 (1984)
        # eq. (33) of Ref. [1]_
        Ntheta = 1.0 + 2.8343 * theta ** 2 - 0.2151 * theta ** 3 + 5.2759 * theta ** 4
        # eq. (34) of Ref. [1]_
        Dtheta = 1.0 + 3.9431 * theta ** 2 + 7.9138 * theta ** 4
        # eq. (32) of Ref. [1]_
        h = Ntheta / Dtheta * np.tanh(1.0 / theta)
        # grad h(x)
        gradh = -(Ntheta / Dtheta) / np.cosh(1 / theta) ** 2 / (theta ** 2) - np.tanh(  # derivative of tanh(1/x)
            1.0 / theta
        ) * (
            Ntheta * (7.8862 * theta + 31.6552 * theta ** 3) / Dtheta ** 2  # derivative of 1/Dtheta
            + (5.6686 * theta - 0.6453 * theta ** 2 + 21.1036 * theta ** 3) / Dtheta
        )  # derivative of Ntheta
        # eq.(31) of Ref. [1]_
        b = 1.0 - 2.0 / (8.0 * (params.kF * params.lambda_TF) ** 2) * (h - 2.0 * theta * gradh)
    else:
        b = 1.0

    params.b = b

    # Monotonic decay
    if params.nu <= 1:
        # eq. (29) of Ref. [1]_
        params.lambda_p = params.lambda_TF * np.sqrt(params.nu / (2.0 * b + 2.0 * np.sqrt(b ** 2 - params.nu)))
        params.lambda_m = params.lambda_TF * np.sqrt(params.nu / (2.0 * b - 2.0 * np.sqrt(b ** 2 - params.nu)))
        params.alpha = b / np.sqrt(b - params.nu)

    # Oscillatory behavior
    if params.nu > 1:
        # eq. (29) of Ref. [1]_
        params.gamma_m = params.lambda_TF * np.sqrt(params.nu / (np.sqrt(params.nu) - b))
        params.gamma_p = params.lambda_TF * np.sqrt(params.nu / (np.sqrt(params.nu) + b))
        params.alphap = b / np.sqrt(params.nu - b)

    potential.matrix = np.zeros((7, params.num_species, params.num_species))

    potential.matrix[1, :, :] = params.nu

    for i, q1 in enumerate(params.species_charges):

        for j, q2 in enumerate(params.species_charges):

            if params.nu <= 1:
                potential.matrix[0, i, j] = q1 * q2 / (2.0 * params.fourpie0)
                potential.matrix[2, i, j] = 1.0 + params.alpha
                potential.matrix[3, i, j] = 1.0 - params.alpha
                potential.matrix[4, i, j] = 1.0 / params.lambda_m
                potential.matrix[5, i, j] = 1.0 / params.lambda_p

            if params.nu > 1:
                potential.matrix[0, i, j] = q1 * q2 / params.fourpie0
                potential.matrix[2, i, j] = 1.0
                potential.matrix[3, i, j] = params.alphap
                potential.matrix[4, i, j] = 1.0 / params.gamma_m
                potential.matrix[5, i, j] = 1.0 / params.gamma_p

    potential.matrix[6, :, :] = potential.a_rs

    if potential.method == "pppm":
        raise AlgorithmError("pppm algorithm not implemented yet.")

    potential.force = egs_force
    # EGS is always smaller than pure Yukawa.
    # Therefore the force error is chosen to be the same as Yukawa's.
    # This overestimates it, but it doesn't matter.

    # The rescaling constant is sqrt ( na^4 ) = sqrt( 3 a/(4pi) )
    params.force_error = force_error_analytic_pp(
        potential.type, potential.rc, potential.matrix, np.sqrt(3.0 * params.a_ws / (4.0 * np.pi))
    )


@njit
def egs_force(r_in, pot_matrix):
    """
    Calculates Potential and force between particles using the EGS Potential.

    Parameters
    ----------
    r_in : float
        Particles' distance.

    pot_matrix : array
        EGS potential parameters. \n
        Shape = (6, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Return
    ------
    U : float
        Potential.

    fr : float
        Force.

    """

    rs = pot_matrix[6]
    # Branchless programming
    r = r_in * (r_in >= rs) + rs * (r_in < rs)

    # nu = pot_matrix[1]
    if pot_matrix[1] <= 1.0:
        temp1 = pot_matrix[2] * np.exp(-r * pot_matrix[4])
        temp2 = pot_matrix[3] * np.exp(-r * pot_matrix[5])
        # Potential
        U = (temp1 + temp2) * pot_matrix[0] / r
        # Force
        fr = U / r + pot_matrix[0] * (temp1 * pot_matrix[4] + temp2 * pot_matrix[5]) / r

    else:
        cos = np.cos(r * pot_matrix[4])
        sin = np.sin(r * pot_matrix[4])
        exp = pot_matrix[0] * np.exp(-r * pot_matrix[5])
        U = (cos + pot_matrix[3] * sin) * exp / r
        fr = U / r  # derivative of 1/r
        fr += U * pot_matrix[5]  # derivative of exp
        fr += pot_matrix[4] * (sin - pot_matrix[3] * cos) * exp / r

    return U, fr



"""Module of mathematical functions."""

import numpy as np
import numba as nb
import scipy.signal as scp_signal

TWOPI = 2.0 * np.pi


def correlationfunction(At, Bt):
    """
    Calculate the correlation function between :math:`\\mathbf{A}(t)` and :math:`\\mathbf{B}(t)` using
    :func:`scipy.signal.correlate`

    .. math::
        C_{AB}(\\tau) =  \\sum_j^D \\sum_i^T A_j(t_i)B_j(t_i + \\tau)

    where :math:`D` is the number of dimensions and :math:`T` is the total length
    of the simulation.

    Parameters
    ----------
    At : numpy.ndarray
        Observable to correlate.

    Bt : numpy.ndarray
        Observable to correlate.

    Returns
    -------
    full_corr : numpy.ndarray
        Correlation function :math:`C_{AB}(\\tau)`
    """
    no_steps = At.size

    # Calculate the full correlation function.
    full_corr = scp_signal.correlate(At, Bt, mode="full")
    # Normalization of the full correlation function, Similar to norm_counter
    norm_corr = np.array([no_steps - ii for ii in range(no_steps)])
    # Find the mid point of the array
    mid = full_corr.size // 2
    # I want only the second half of the array, i.e. the positive lags only
    return full_corr[mid:] / norm_corr


@nb.njit
def fast_integral_loop(time, integrand):
    """Compute the integral with varying upper limit.

    Parameters
    ----------
    time: numpy.ndarray
        Domain of integration.

    integrand: numpy.ndarray
        Integrand.

    Returns
    -------
    integral : numpy.ndarray
        Integral with increasing upper limit. Shape = (time.len()).


    """
    integral = np.zeros_like(integrand)
    for it in range(1, len(time)):
        integral[it] = np.trapz(integrand[:it], x=time[:it])

    return integral


@nb.njit
def yukawa_green_function(k, alpha, kappa):
    """
    Evaluate the Green's function of Coulomb/Yukawa potential.

    .. math::

        G(k) = \\frac{4 \\pi }{\\kappa^2 + k^2} e^{- (k^2 + \\kappa^2)/4 \\alpha^2 }

    Parameters
    ----------
    k : numpy.ndarray, float
        Range or value at which to calculate the function.

    alpha: float
        Ewald screening parameter.

    kappa : float
        Inverse screening length.

    Returns
    -------
    _ : numpy.ndarray, float
        Green's function. See equation above

    """
    return 4.0 * np.pi * np.exp(-(k ** 2 + kappa ** 2) / (2 * alpha) ** 2) / (kappa ** 2 + k ** 2)


@nb.njit
def betamp(m, p, alpha, kappa):
    """
    Calculate the integral of the Yukawa Green's function

    .. math::

        \\beta(p,m) = \\int_0^\\infty dk \\, G_k^2 k^{2 (p + m + 2)}.

    See eq.(37) in :cite:`Dharuman2017`.
    """
    xa = np.linspace(0.0001, 500, 5000)
    Gk = yukawa_green_function(xa, alpha, kappa)
    return np.trapz(Gk * Gk * xa ** (2 * (m + p + 2)), x=xa)


@nb.njit
def force_error_approx_pppm(kappa, rc, p, h, alpha):
    """
    Calculate the total force error for a given value of the PPPM parameters.

    Parameters
    ----------
    kappa : float
        Inverse screening length.

    rc : float
        Cutoff length.

    p : int
        Charge assignment order.

    h : float
        Distance between two mesh points. Same for all directions.

    alpha : float
        Ewald screening parameter.

    Returns
    -------

    Tot_Delta_F : float
        Total force error given by

        .. math::
            \\Delta F = \\sqrt{\\Delta F_{\\textrm {pp}}^2 + \\Delta F_{\\textrm {pm}}^2 }

    pp_force_error : float
        PP force error.

    pm_force_error: float
        PM force error.

    """
    # Coefficients from :cite:`Deserno1998`
    if p == 1:
        Cmp = np.array([2 / 3])
    elif p == 2:
        Cmp = np.array([2 / 45, 8 / 189])
    elif p == 3:
        Cmp = np.array([4 / 495, 2 / 225, 8 / 1485])
    elif p == 4:
        Cmp = np.array([2 / 4725, 16 / 10395, 5528 / 3869775, 32 / 42525])
    elif p == 5:
        Cmp = np.array([4 / 93555, 2764 / 11609325, 8 / 25515, 7234 / 32531625, 350936 / 3206852775])
    elif p == 6:
        Cmp = np.array(
            [
                2764 / 638512875,
                16 / 467775,
                7234 / 119282625,
                1403744 / 25196700375,
                1396888 / 40521009375,
                2485856 / 152506344375,
            ]
        )
    elif p == 7:
        Cmp = np.array(
            [
                8 / 18243225,
                7234 / 1550674125,
                701872 / 65511420975,
                2793776 / 225759909375,
                1242928 / 132172165125,
                1890912728 / 352985880121875,
                21053792 / 8533724574375,
            ]
        )

    somma = 0.0
    for m in np.arange(p):
        expp = 2 * (m + p)
        somma += Cmp[m] * (2 / (1 + expp)) * betamp(m, p, alpha, kappa) * (h / 2.0) ** expp
    # eq.(36) in :cite:`Dharuman2017`
    pm_force_error = np.sqrt(3.0 * somma) / (2.0 * np.pi)

    # eq.(30) from :cite:`Dharuman2017`
    pp_force_error = 2.0 * np.exp(-((0.5 * kappa / alpha) ** 2) - alpha ** 2 * rc ** 2) / np.sqrt(rc)
    # eq.(42) from :cite:`Dharuman2017`
    Tot_DeltaF = np.sqrt(pm_force_error ** 2 + pp_force_error ** 2)

    return Tot_DeltaF, pp_force_error, pm_force_error


@nb.njit
def force_error_analytic_pp(potential_type, cutoff_length, potential_matrix, rescaling_const):
    """
    Calculate the force error from its analytic formula.

    Parameters
    ----------
    potential_type: str
        Type of potential used.

    potential_matrix: numpy.ndarray
        Potential parameters.

    cutoff_length : float, numpy.ndarray
        Cutoff radius of the potential.

    rescaling_const: float
        Constant by which to rescale the force error.

    Returns
    -------
    force_error: float
        Force error in units of Q^2/(4pi eps0) 1/a^2.

    """

    if potential_type in ["yukawa", "egs", "qsp"]:
        force_error = np.sqrt(TWOPI * potential_matrix[1, 0, 0]) * np.exp(-cutoff_length * potential_matrix[1, 0, 0])
    elif potential_type == "moliere":
        # Choose the smallest screening length for force error calculation

        force_error = np.sqrt(TWOPI * potential_matrix[:, 0, 0].min()) * np.exp(
            -cutoff_length * potential_matrix[1, 0, 0]
        )

    elif potential_type == "lj":
        # choose the highest sigma in case of multispecies
        sigma = potential_matrix[1, :, :].max()
        high_pow = potential_matrix[2, 0, 0]
        exponent = 2 * high_pow - 1
        force_error_tmp = high_pow ** 2 * sigma ** (2 * high_pow) / cutoff_length ** exponent
        force_error_tmp /= exponent
        force_error = np.sqrt(force_error_tmp)

    # Renormalize
    force_error *= rescaling_const

    return force_error



"""
Module for handling Hard-Sphere Yukawa potential.

Potential
*********

The Hard-Sphere Yukawa potential between two charges :math:`q_i` and :math:`q_j` at distant :math:`r` is defined as

.. math::
    U_{ab}(r) = \\left ( \\frac{\\sigma}{r} \\right )^{n} \\frac{q_a q_b}{4 \\pi \\epsilon_0} \\frac{e^{- \\kappa r} }{r}.

where :math:`\\kappa = 1/\\lambda` is the screening parameter.

Potential Attributes
********************

The elements of the :attr:`sarkas.potentials.core.Potential.pot_matrix` are:

.. code-block:: python

    pot_matrix[0] = q_iq_j^2/(4 pi eps0)
    pot_matrix[1] = 1/lambda
    pot_matrix[2] = Ewald screening parameter
    pot_matrix[3] = short range cutoff
"""

import numpy as np
from numba import njit
from sarkas.utilities.maths import force_error_analytic_pp


@njit
def hs_yukawa_force(r, pot_matrix):
    """
    Calculates Potential and Force between two particles.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : numpy.ndarray
        It contains potential dependent variables. \n
        Shape = (3, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    U : float
        Potential.

    force : float
        Force between two particles.

    """
    U_y = pot_matrix[0] * np.exp(-pot_matrix[1] * r) / r
    U_hs = (pot_matrix[2] / r) ** (50)
    U = U_y + U_hs
    force = U_y * (1.0 / r + pot_matrix[1]) + (50.0) * U_hs / r

    return U, force


@njit
def force_deriv(r, pot_matrix):
    """Calculate the second derivative of the potential.

    Parameters
    ----------

    r : float
        Distance between particles

    pot_matrix : numpy.ndarray
        Values of the potential constants. \n
        Shape = (3, :attr:`sarkas.core.Parameters.num_species`, :attr:`sarkas.core.Parameters.num_species`)

    Returns
    -------
    f_dev : float
        Second derivative of potential.

    """
    kappa_r = pot_matrix[1] * r
    U2 = pot_matrix[0] * np.exp(-kappa_r) / r ** 3
    f_dev = U2 * (2.0 * (1.0 + kappa_r) + kappa_r ** 2)
    return f_dev


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params: sarkas.core.Parameters
        Simulation's parameters.

    """

    potential.matrix = np.zeros((3, params.num_species, params.num_species))
    potential.matrix[1, :, :] = 1.0 / params.lambda_TF

    for i, q1 in enumerate(params.species_charges):
        for j, q2 in enumerate(params.species_charges):
            potential.matrix[0, i, j] = q1 * q2 / params.fourpie0

    potential.matrix[2, :, :] = potential.hs_radius

    if potential.method == "pp":
        # The rescaling constant is sqrt ( na^4 ) = sqrt( 3 a/(4pi) )
        potential.force = hs_yukawa_force
        params.force_error = force_error_analytic_pp(
            "yukawa", potential.rc, potential.matrix, np.sqrt(3.0 * params.a_ws / (4.0 * np.pi))
        )
        # # Force error calculated from eq.(43) in Ref.[1]_
        # params.force_error = np.sqrt( TWOPI / params.lambda_TF) * np.exp(- potential.rc / params.lambda_TF)
        # # Renormalize
        # params.force_error *= params.a_ws ** 2 * np.sqrt(params.total_num_ptcls / params.pbox_volume)
    elif potential.method == "pppm":
        raise ValueError("PPPM algorithm not supported.")



import os
import sys
import re
import yaml
import csv
import pickle
import numpy as np
from pyfiglet import print_figlet, Figlet
from IPython import get_ipython

if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
    # If you are using Jupyter Notebook
    from tqdm import tqdm_notebook as tqdm
else:
    # If you are using IPython or Python kernel
    from tqdm import tqdm

FONTS = ["speed", "starwars", "graffiti", "chunky", "epic", "larry3d", "ogre"]

# Light Colors.
LIGHT_COLORS = [
    "255;255;255",
    "13;177;75",
    "153;162;162",
    "240;133;33",
    "144;154;183",
    "209;222;63",
    "232;217;181",
    "200;154;88",
    "148;174;74",
    "203;90;40",
]

# Dark Colors.
DARK_COLORS = ["24;69;49", "0;129;131", "83;80;84", "110;0;95"]


class InputOutput:
    def __init__(self, process: str = None):
        """Set default directory names."""
        self.process = process if process else "preprocessing"
        self.input_file = None
        self.equilibration_dir = "Equilibration"
        self.production_dir = "Production"
        self.magnetization_dir = "Magnetization"
        self.simulations_dir = "Simulations"
        self.processes_dir = None
        self.simulation_dir = "Simulation"
        self.preprocessing_dir = "PreProcessing"
        self.postprocessing_dir = "PostProcessing"
        self.prod_dump_dir = "dumps"
        self.eq_dump_dir = "dumps"
        self.mag_dump_dir = "dumps"
        self.job_dir = None
        self.job_id = None
        self.log_file = None
        self.preprocess_file = None
        self.preprocessing = False
        self.magnetized = False
        self.electrostatic_equilibration = False
        self.verbose = False
        self.xyz_dir = None
        self.xyz_filename = None

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = "InputOuput( \n"
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ")"
        return disp

    def from_dict(self, input_dict: dict):
        """
        Update attributes from input dictionary.

        Parameters
        ----------
        input_dict: dict
            Dictionary to be copied.

        """
        self.__dict__.update(input_dict)

    def setup(self):
        """Create file paths and directories for the simulation."""
        self.create_file_paths()
        self.make_directories()
        self.file_header()

    def from_yaml(self, filename):
        """
        Parse inputs from YAML file.

        Parameters
        ----------
        filename: str
            Input YAML file.

        Returns
        -------
        dics : dict
            Content of YAML file parsed in a nested dictionary
        """

        self.input_file = filename
        with open(filename, "r") as stream:
            dics = yaml.load(stream, Loader=yaml.FullLoader)
            self.__dict__.update(dics["IO"])

        if "Parameters" in dics.keys():
            keyed = "Parameters"
            for key, value in dics[keyed].items():

                if key == "verbose":
                    self.verbose = value

                if key == "magnetized":
                    self.magnetized = value

                if key == "load_method":
                    self.load_method = value
                    if value[-7:] == "restart":
                        self.restart = True
                    else:
                        self.restart = False

                if key == "preprocessing":
                    self.preprocessing = value

        if "Integrator" in dics.keys():
            keyed = "Integrator"
            for key, value in dics[keyed].items():

                if key == "electrostatic_equilibration":
                    self.electrostatic_equilibration = value

        # rdf_nbins can be defined in either Parameters or Postprocessing. However, Postprocessing will always
        # supersede Parameters choice.
        if "Observables" in dics.keys():
            for i in dics["Observables"]:
                if "RadialDistributionFunction" in i.keys():
                    dics["Parameters"]["rdf_nbins"] = i["RadialDistributionFunction"]["no_bins"]

        return dics

    def create_file_paths(self):
        """Create all directories', subdirectories', and files' paths ."""

        if self.job_dir is None:
            self.job_dir = os.path.basename(self.input_file).split(".")[0]

        if self.job_id is None:
            self.job_id = self.job_dir

        self.job_dir = os.path.join(self.simulations_dir, self.job_dir)

        # Create Processes directories
        self.processes_dir = [
            os.path.join(self.job_dir, self.preprocessing_dir),
            os.path.join(self.job_dir, self.simulation_dir),
            os.path.join(self.job_dir, self.postprocessing_dir),
        ]

        # Redundancy
        self.preprocessing_dir = self.processes_dir[0]
        self.simulation_dir = self.processes_dir[1]
        self.postprocessing_dir = self.processes_dir[2]

        # Redirect to the correct process folder
        if self.process == "preprocessing":
            indx = 0
        else:
            # Note that Postprocessing needs the link to simulation's folder
            # because that is where I look for energy files and pickle files
            indx = 1

        # Equilibration directory and sub_dir
        self.equilibration_dir = os.path.join(self.processes_dir[indx], self.equilibration_dir)
        self.eq_dump_dir = os.path.join(self.equilibration_dir, "dumps")
        # Production dir and sub_dir
        self.production_dir = os.path.join(self.processes_dir[indx], self.production_dir)
        self.prod_dump_dir = os.path.join(self.production_dir, "dumps")

        # Production phase filenames
        self.prod_energy_filename = os.path.join(self.production_dir, "ProductionEnergy_" + self.job_id + ".csv")
        self.prod_ptcls_filename = os.path.join(self.prod_dump_dir, "checkpoint_")

        # Equilibration phase filenames
        self.eq_energy_filename = os.path.join(self.equilibration_dir, "EquilibrationEnergy_" + self.job_id + ".csv")
        self.eq_ptcls_filename = os.path.join(self.eq_dump_dir, "checkpoint_")

        # Magnetic dir
        if self.electrostatic_equilibration:
            self.magnetization_dir = os.path.join(self.processes_dir[indx], self.magnetization_dir)
            self.mag_dump_dir = os.path.join(self.magnetization_dir, "dumps")
            # Magnetization phase filenames
            self.mag_energy_filename = os.path.join(self.magnetization_dir, "MagnetizationEnergy_" + self.job_id + ".csv")
            self.mag_ptcls_filename = os.path.join(self.mag_dump_dir, "checkpoint_")

        if self.process == "postprocessing":
            indx = 2  # Redirect to the correct folder

        # Log File
        if self.log_file is None:
            self.log_file = os.path.join(self.processes_dir[indx], "log_" + self.job_id + ".out")
        else:
            self.log_file = os.path.join(self.processes_dir[indx], self.log_file)

    def make_directories(self):
        """Create directories if non-existent."""

        # Check if the directories exist
        if not os.path.exists(self.simulations_dir):
            os.mkdir(self.simulations_dir)

        if not os.path.exists(self.job_dir):
            os.mkdir(self.job_dir)

        # Create Process' directories and their subdir
        for i in self.processes_dir:
            if not os.path.exists(i):
                os.mkdir(i)
        # The following automatically create directories in the correct Process
        if not os.path.exists(self.equilibration_dir):
            os.mkdir(self.equilibration_dir)

        if not os.path.exists(self.eq_dump_dir):
            os.mkdir(self.eq_dump_dir)

        if not os.path.exists(self.production_dir):
            os.mkdir(self.production_dir)

        if not os.path.exists(self.prod_dump_dir):
            os.mkdir(self.prod_dump_dir)

        if self.electrostatic_equilibration:
            if not os.path.exists(self.magnetization_dir):
                os.mkdir(self.magnetization_dir)

            if not os.path.exists(self.mag_dump_dir):
                os.mkdir(self.mag_dump_dir)

        if self.preprocessing:
            if not os.path.exists(self.preprocessing_dir):
                os.mkdir(self.preprocessing_dir)

        if not os.path.exists(self.postprocessing_dir):
            os.mkdir(self.postprocessing_dir)

    def file_header(self):
        """Create the log file and print the figlet if not a restart run."""

        if not self.restart:
            with open(self.log_file, "w+") as f_log:
                figlet_obj = Figlet(font="starwars")
                print(figlet_obj.renderText("Sarkas"), file=f_log)
                print("An open-source pure-Python molecular dynamics suite for non-ideal plasmas.", file=f_log)

        # Print figlet to screen if verbose
        if self.verbose:
            self.screen_figlet()

    def simulation_summary(self, simulation):
        """
        Print out to file a summary of simulation's parameters.
        If verbose output then it will print twice: the first time to file and second time to screen.

        Parameters
        ----------
        simulation : sarkas.processes.Process
            Simulation's parameters

        """

        screen = sys.stdout
        f_log = open(self.log_file, "a+")

        repeat = 2 if self.verbose else 1
        # redirect printing to file
        sys.stdout = f_log

        # Print to file first then to screen if repeat == 2
        while repeat > 0:

            if simulation.parameters.load_method in ["production_restart", "prod_restart"]:
                print("\n\n--------------------------- Production Restart -------------------------------------")
                self.time_info(simulation)
            elif simulation.parameters.load_method in ["equilibration_restart", "eq_restart"]:
                print("\n\n------------------------ Equilibration Restart ----------------------------------")
                self.time_info(simulation)
            elif simulation.parameters.load_method in ["magnetization_restart", "mag_restart"]:
                print("\n\n------------------------ Magnetization Restart ----------------------------------")
                self.time_info(simulation)
            elif self.process == "postprocessing":
                # Header of process
                process_title = "{:^80}".format(self.process.capitalize())
                print("\n\n")
                print(*["*" for i in range(50)])
                print(process_title)
                print(*["*" for i in range(50)])

                print("\nJob ID: ", self.job_id)
                print("Job directory: ", self.job_dir)
                print("PostProcessing directory: \n", self.postprocessing_dir)

                print("\nEquilibration dumps directory: ", self.eq_dump_dir)
                print("Production dumps directory: \n", self.prod_dump_dir)

                print("\nEquilibration Thermodynamics file: \n", self.eq_energy_filename)
                print("Production Thermodynamics file: \n", self.prod_energy_filename)

            else:

                # Header of process
                process_title = "{:^80}".format(self.process.capitalize())
                print("\n\n")
                print(*["*" for i in range(50)])
                print(process_title)
                print(*["*" for i in range(50)])

                print("\nJob ID: ", self.job_id)
                print("Job directory: ", self.job_dir)
                print("\nEquilibration dumps directory: \n", self.eq_dump_dir)
                print("Production dumps directory: \n", self.prod_dump_dir)

                print("\nEquilibration Thermodynamics file: \n", self.eq_energy_filename)
                print("Production Thermodynamics file: \n", self.prod_energy_filename)

                if hasattr(simulation.parameters, "rand_seed"):
                    print("Random Seed = ", simulation.parameters.rand_seed)

                print("\nPARTICLES:")
                print("Total No. of particles = ", simulation.parameters.total_num_ptcls)
                print("No. of species = ", len(simulation.species))
                for isp, sp in enumerate(simulation.species):
                    print("Species ID: {}".format(isp))
                    sp.pretty_print(simulation.potential.type, simulation.parameters.units)

                simulation.parameters.pretty_print()

                print("\nPOTENTIAL: ", simulation.potential.type)
                self.potential_info(simulation)

                print("\nALGORITHM: ", simulation.potential.method)
                self.algorithm_info(simulation)

                print("\nTHERMOSTAT: ")
                simulation.thermostat.pretty_print()

                print("\nINTEGRATOR: ")
                simulation.integrator.pretty_print(
                    simulation.parameters.total_plasma_frequency,
                    simulation.parameters.load_method,
                    simulation.parameters.restart_step,
                )

            repeat -= 1
            sys.stdout = screen  # Restore the original sys.stdout

        f_log.close()

    def time_stamp(self, time_stamp, t):
        """
        Print out to screen elapsed times. If verbose output, print to file first and then to screen.

        Parameters
        ----------
        time_stamp : str
            Array of time stamps.

        t : float
            Elapsed time.
        """
        screen = sys.stdout
        f_log = open(self.log_file, "a+")
        repeat = 2 if self.verbose else 1
        t_hrs, t_min, t_sec, t_msec, t_usec, t_nsec = t
        # redirect printing to file
        sys.stdout = f_log

        while repeat > 0:
            if "Potential Initialization" in time_stamp:
                print("\n\n{:-^70} \n".format(" Initialization Times "))
            if t_hrs == 0 and t_min == 0 and t_sec <= 2:
                print(
                    "\n{} Time: {} sec {} msec {} usec {} nsec".format(
                        time_stamp, int(t_sec), int(t_msec), int(t_usec), int(t_nsec)
                    )
                )
            else:
                print("\n{} Time: {} hrs {} min {} sec".format(time_stamp, int(t_hrs), int(t_min), int(t_sec)))

            repeat -= 1
            sys.stdout = screen

        f_log.close()

    def timing_study(self, simulation):
        """
        Info specific for timing study.

        Parameters
        ----------
        simulation: sarkas.core.Simulation
            Simulation's parameters

        """
        screen = sys.stdout
        f_log = open(self.log_file, "a+")
        repeat = 2 if self.verbose else 1

        # redirect printing to file
        sys.stdout = f_log

        # Print to file first then to screen if repeat == 2
        while repeat > 0:
            print("\n\n------------ Conclusion ------------\n")
            print("Suggested Mesh = [ {} , {} , {} ]".format(*simulation.potential.pppm_mesh))
            print(
                "Suggested Ewald parameter alpha = {:2.4f} / a_ws = {:1.6e} ".format(
                    simulation.potential.pppm_alpha_ewald * simulation.parameters.a_ws,
                    simulation.potential.pppm_alpha_ewald,
                ),
                end="",
            )
            print("[1/cm]" if simulation.parameters.units == "cgs" else "[1/m]")
            print(
                "Suggested rcut = {:2.4f} a_ws = {:.6e} ".format(
                    simulation.potential.rc / simulation.parameters.a_ws, simulation.potential.rc
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")

            self.algorithm_info(simulation)
            repeat -= 1
            sys.stdout = screen  # Restore the original sys.stdout

        f_log.close()

    def preprocess_sizing(self, sizes):
        """Print the estimated file sizes. """

        screen = sys.stdout
        f_log = open(self.log_file, "a+")
        repeat = 2 if self.verbose else 1

        # redirect printing to file
        sys.stdout = f_log
        while repeat > 0:
            print("\n\n{:=^70} \n".format(" Filesize Estimates "))
            size_GB, size_MB, size_KB, rem = convert_bytes(sizes[0, 0])
            print("\nEquilibration:\n")
            print(
                "Checkpoint filesize: {} GB {} MB {} KB {} bytes".format(
                    int(size_GB), int(size_MB), int(size_KB), int(rem)
                )
            )

            size_GB, size_MB, size_KB, rem = convert_bytes(sizes[0, 1])
            print(
                "Checkpoint folder size: {} GB {} MB {} KB {} bytes".format(
                    int(size_GB), int(size_MB), int(size_KB), int(rem)
                )
            )
            if self.electrostatic_equilibration:
                print("\nMagnetization:\n")
                size_GB, size_MB, size_KB, rem = convert_bytes(sizes[2, 0])
                print(
                    "Checkpoint filesize: {} GB {} MB {} KB {} bytes".format(
                        int(size_GB), int(size_MB), int(size_KB), int(rem)
                    )
                )

                size_GB, size_MB, size_KB, rem = convert_bytes(sizes[2, 1])
                print(
                    "Checkpoint folder size: {} GB {} MB {} KB {} bytes".format(
                        int(size_GB), int(size_MB), int(size_KB), int(rem)
                    )
                )

            size_GB, size_MB, size_KB, rem = convert_bytes(sizes[1, 0])
            print("\nProduction:\n")
            print(
                "Checkpoint filesize: {} GB {} MB {} KB {} bytes".format(
                    int(size_GB), int(size_MB), int(size_KB), int(rem)
                )
            )

            size_GB, size_MB, size_KB, rem = convert_bytes(sizes[1, 1])
            print(
                "Checkpoint folder size: {} GB {} MB {} KB {} bytes".format(
                    int(size_GB), int(size_MB), int(size_KB), int(rem)
                )
            )

            size_GB, size_MB, size_KB, rem = convert_bytes(sizes[:, 1].sum())
            print(
                "\nTotal minimum needed space: {} GB {} MB {} KB {} bytes".format(
                    int(size_GB), int(size_MB), int(size_KB), int(rem)
                )
            )
            repeat -= 1
            sys.stdout = screen

        f_log.close()

    def preprocess_timing(self, str_id, t, loops):
        """Print times estimates of simulation to file first and then to screen if verbose."""
        screen = sys.stdout
        f_log = open(self.log_file, "a+")
        repeat = 2 if self.verbose else 1
        t_hrs, t_min, t_sec, t_msec, t_usec, t_nsec = t
        # redirect printing to file
        sys.stdout = f_log
        while repeat > 0:
            if str_id == "header":
                print("\n\n{:=^70} \n".format(" Times Estimates "))
            elif str_id == "GF":
                print(
                    "Optimal Green's Function Time: \n"
                    "{} min {} sec {} msec {} usec {} nsec \n".format(
                        int(t_min), int(t_sec), int(t_msec), int(t_usec), int(t_nsec)
                    )
                )

            elif str_id == "PP":
                print(
                    "Time of PP acceleration calculation averaged over {} steps: \n"
                    "{} min {} sec {} msec {} usec {} nsec \n".format(
                        loops - 1, int(t_min), int(t_sec), int(t_msec), int(t_usec), int(t_nsec)
                    )
                )

            elif str_id == "PM":
                print(
                    "Time of PM acceleration calculation averaged over {} steps: \n"
                    "{} min {} sec {} msec {} usec {} nsec \n".format(
                        loops - 1, int(t_min), int(t_sec), int(t_msec), int(t_usec), int(t_nsec)
                    )
                )

            elif str_id == "Equilibration":
                print(
                    "Time of a single equilibration step averaged over {} steps: \n"
                    "{} min {} sec {} msec {} usec {} nsec \n".format(
                        loops - 1, int(t_min), int(t_sec), int(t_msec), int(t_usec), int(t_nsec)
                    )
                )
            elif str_id == "Magnetization":
                print(
                    "Time of a single magnetization step averaged over {} steps: \n"
                    "{} min {} sec {} msec {} usec {} nsec \n".format(
                        loops - 1, int(t_min), int(t_sec), int(t_msec), int(t_usec), int(t_nsec)
                    )
                )
            elif str_id == "Production":
                print(
                    "Time of a single production step averaged over {} steps: \n"
                    "{} min {} sec {} msec {} usec {} nsec \n".format(
                        loops - 1, int(t_min), int(t_sec), int(t_msec), int(t_usec), int(t_nsec)
                    )
                )

                print("\n\n{:-^70} \n".format(" Total Estimated Times "))
            repeat -= 1
            sys.stdout = screen

        f_log.close()

    def postprocess_info(self, simulation, write_to_file=False, observable=None):
        """
        Print Post-processing info to file and/or screen in a reader-friendly format.

        Parameters
        ----------
        simulation : sarkas.processes.PostProcess
            Sarkas processing stage.

        write_to_file : bool
            Flag for printing info also to file. Default= False.

        observable : str
            Observable whose info to print. Default = None.
            Choices = ['header','rdf', 'ccf', 'dsf', 'ssf', 'vm']

        """

        choices = ["header", "rdf", "ccf", "dsf", "ssf", "vd"]
        msg = (
            "Observable not defined. \n "
            "Please choose an observable from this list \n"
            "'rdf' = Radial Distribution Function, \n"
            "'ccf' = Current Correlation Function, \n"
            "'dsf' = Dynamic Structure Function, \n"
            "'ssf' = Static Structure Factor, \n"
            "'vd' = Velocity Distribution"
        )
        if observable is None:
            raise ValueError(msg)

        if observable not in choices:
            raise ValueError(msg)

        if write_to_file:
            screen = sys.stdout
            f_log = open(self.log_file, "a+")
            repeat = 2 if self.verbose else 1

            # redirect printing to file
            sys.stdout = f_log
        else:
            repeat = 1

        while repeat > 0:
            if observable == "header":
                # Header of process
                process_title = "{:^80}".format(self.process.capitalize())
                print("\n\n")
                print(*["*" for i in range(50)])
                print(process_title)
                print(*["*" for i in range(50)])

            elif observable == "rdf":
                simulation.rdf.pretty_print()
            elif observable == "ssf":
                simulation.ssf.pretty_print()
            elif observable == "dsf":
                simulation.dsf.pretty_print()
            elif observable == "ccf":
                simulation.ccf.pretty_print()
            elif observable == "vd":
                simulation.vm.setup(simulation.parameters)
                print("\nVelocity Moments:")
                print("Maximum no. of moments = {}".format(simulation.vm.max_no_moment))
                print("Maximum velocity moment = {}".format(int(2 * simulation.vm.max_no_moment)))

            repeat -= 1

            if write_to_file:
                sys.stdout = screen

        if write_to_file:
            f_log.close()

    @staticmethod
    def screen_figlet():
        """
        Print a colored figlet of Sarkas to screen.
        """
        if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
            # Assume white background in Jupyter Notebook
            clr = DARK_COLORS[np.random.randint(0, len(DARK_COLORS))]
        else:
            # Assume dark background in IPython/Python Kernel
            clr = LIGHT_COLORS[np.random.randint(0, len(LIGHT_COLORS))]
        fnt = FONTS[np.random.randint(0, len(FONTS))]
        print_figlet("\nSarkas\n", font=fnt, colors=clr)

        print("\nAn open-source pure-python molecular dynamics suite for non-ideal plasmas.\n\n")

    @staticmethod
    def time_info(simulation):
        """
        Print time simulation's parameters.

        Parameters
        ----------
        simulation: sarkas.core.Simulation
            Simulation's parameters.

        """
        wp_dt = simulation.parameters.total_plasma_frequency * simulation.integrator.dt
        print("Time step = {:.6e} [s]".format(simulation.integrator.dt))
        if simulation.potential.type in ["Yukawa", "EGS", "Coulomb", "Moliere"]:
            print("Total plasma frequency = {:1.6e} [rad/s]".format(simulation.parameters.total_plasma_frequency))
            print("w_p dt = {:2.4f}".format(wp_dt))
            if simulation.parameters.magnetized:
                if simulation.parameters.num_species > 1:
                    high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
                    low_wc_dt = simulation.parameters.species_cyclotron_frequencies.min() * simulation.integrator.dt
                    print("Highest w_c dt = {:2.4f}".format(high_wc_dt))
                    print("Smalles w_c dt = {:2.4f}".format(low_wc_dt))
                else:
                    high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
                    print("w_c dt = {:2.4f}".format(high_wc_dt))
        elif simulation.potential.type == "QSP":
            print("e plasma frequency = {:.6e} [rad/s]".format(simulation.species[0].plasma_frequency))
            print("ion plasma frequency = {:.6e} [rad/s]".format(simulation.species[1].plasma_frequency))
            print("w_pe dt = {:2.4f}".format(simulation.integrator.dt * simulation.species[0].plasma_frequency))
            if simulation.parameters.magnetized:
                if simulation.parameters.num_species > 1:
                    high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
                    low_wc_dt = simulation.parameters.species_cyclotron_frequencies.min() * simulation.integrator.dt
                    print("Electron w_ce dt = {:2.4f}".format(high_wc_dt))
                    print("Ions w_ci dt = {:2.4f}".format(low_wc_dt))
                else:
                    high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
                    print("w_c dt = {:2.4f}".format(high_wc_dt))
        elif simulation.potential.type == "LJ":
            print(
                "Total equivalent plasma frequency = {:1.6e} [rad/s]".format(simulation.parameters.total_plasma_frequency)
            )
            print("w_p dt = {:2.4f}".format(wp_dt))
            if simulation.parameters.magnetized:
                if simulation.parameters.num_species > 1:
                    high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
                    low_wc_dt = simulation.parameters.species_cyclotron_frequencies.min() * simulation.integrator.dt
                    print("Highest w_c dt = {:2.4f}".format(high_wc_dt))
                    print("Smalles w_c dt = {:2.4f}".format(low_wc_dt))
                else:
                    high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
                    print("w_c dt = {:2.4f}".format(high_wc_dt))

        # Print Time steps information
        # Check for restart simulations
        if simulation.parameters.load_method in ["production_restart", "prod_restart"]:
            print("Restart step: {}".format(simulation.parameters.restart_step))
            print(
                "Total production steps = {} \n"
                "Total production time = {:.4e} [s] ~ {} w_p T_prod ".format(
                    simulation.integrator.production_steps,
                    simulation.integrator.production_steps * simulation.integrator.dt,
                    int(simulation.integrator.production_steps * wp_dt),
                )
            )
            print(
                "snapshot interval step = {} \n"
                "snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap".format(
                    simulation.integrator.prod_dump_step,
                    simulation.integrator.prod_dump_step * simulation.integrator.dt,
                    simulation.integrator.prod_dump_step * wp_dt,
                )
            )

        elif simulation.parameters.load_method in ["equilibration_restart", "eq_restart"]:
            print("Restart step: {}".format(simulation.parameters.restart_step))
            print(
                "Total equilibration steps = {} \n"
                "Total equilibration time = {:.4e} [s] ~ {} w_p T_eq".format(
                    simulation.integrator.equilibration_steps,
                    simulation.integrator.equilibration_steps * simulation.integrator.dt,
                    int(simulation.integrator.eq_dump_step * wp_dt),
                )
            )
            print(
                "snapshot interval step = {} \n"
                "snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap".format(
                    simulation.integrator.eq_dump_step,
                    simulation.integrator.eq_dump_step * simulation.integrator.dt,
                    simulation.integrator.eq_dump_step * wp_dt,
                )
            )

        elif simulation.parameters.load_method in ["magnetization_restart", "mag_restart"]:
            print("Restart step: {}".format(simulation.parameters.restart_step))
            print(
                "Total magnetization steps = {} \n"
                "Total magnetization time = {:.4e} [s] ~ {} w_p T_mag".format(
                    simulation.integrator.magnetization_steps,
                    simulation.integrator.magnetization_steps * simulation.integrator.dt,
                    int(simulation.integrator.mag_dump_step * wp_dt),
                )
            )
            print(
                "snapshot interval step = {} \n"
                "snapshot interval time = {:.4e} [s] ~ {:1.4f} w_p T_snap".format(
                    simulation.integrator.mag_dump_step,
                    simulation.integrator.mag_dump_step * simulation.integrator.dt,
                    simulation.integrator.mag_dump_step * wp_dt,
                )
            )
        else:
            # Equilibration
            print(
                "\nEquilibration: \nNo. of equilibration steps = {} \n"
                "Total equilibration time = {:.4e} [s] ~ {} w_p T_eq ".format(
                    simulation.integrator.equilibration_steps,
                    simulation.integrator.equilibration_steps * simulation.integrator.dt,
                    int(simulation.integrator.equilibration_steps * wp_dt),
                )
            )
            print(
                "snapshot interval step = {} \n"
                "snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap".format(
                    simulation.integrator.eq_dump_step,
                    simulation.integrator.eq_dump_step * simulation.integrator.dt,
                    simulation.integrator.eq_dump_step * wp_dt,
                )
            )
            # Magnetization
            if simulation.integrator.electrostatic_equilibration:
                print(
                    "\nMagnetization: \nNo. of magnetization steps = {} \n"
                    "Total magnetization time = {:.4e} [s] ~ {} w_p T_mag ".format(
                        simulation.integrator.magnetization_steps,
                        simulation.integrator.magnetization_stepss * simulation.integrator.dt,
                        int(simulation.integrator.magnetization_steps * wp_dt),
                    )
                )

                print(
                    "snapshot interval step = {} \n"
                    "snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap".format(
                        simulation.integrator.mag_dump_step,
                        simulation.integrator.mag_dump_step * simulation.integrator.dt,
                        simulation.integrator.mag_dump_step * wp_dt,
                    )
                )
            # Production
            print(
                "\nProduction: \nNo. of production steps = {} \n"
                "Total production time = {:.4e} [s] ~ {} w_p T_prod ".format(
                    simulation.integrator.production_steps,
                    simulation.integrator.production_steps * simulation.integrator.dt,
                    int(simulation.integrator.production_steps * wp_dt),
                )
            )
            print(
                "snapshot interval step = {} \n"
                "snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap".format(
                    simulation.integrator.prod_dump_step,
                    simulation.integrator.prod_dump_step * simulation.integrator.dt,
                    simulation.integrator.prod_dump_step * wp_dt,
                )
            )

    @staticmethod
    def algorithm_info(simulation):
        """
        Print algorithm information.

        Parameters
        ----------
        simulation: sarkas.core.Simulation
            Simulation's parameters.


        """
        if simulation.potential.method == "pppm":
            print("Charge assignment order: {}".format(simulation.potential.pppm_cao))
            print("FFT aliases: [{}, {}, {}]".format(*simulation.potential.pppm_aliases))
            print("Mesh: {} x {} x {}".format(*simulation.potential.pppm_mesh))
            print(
                "Ewald parameter alpha = {:2.4f} / a_ws = {:1.6e} ".format(
                    simulation.potential.pppm_alpha_ewald * simulation.parameters.a_ws,
                    simulation.potential.pppm_alpha_ewald,
                ),
                end="",
            )
            print("[1/cm]" if simulation.parameters.units == "cgs" else "[1/m]")
            print(
                "Mesh width = {:.4f}, {:.4f}, {:.4f} a_ws".format(
                    simulation.potential.pppm_h_array[0] / simulation.parameters.a_ws,
                    simulation.potential.pppm_h_array[1] / simulation.parameters.a_ws,
                    simulation.potential.pppm_h_array[2] / simulation.parameters.a_ws,
                )
            )
            print(
                "           = {:.4e}, {:.4e}, {:.4e} ".format(
                    simulation.potential.pppm_h_array[0],
                    simulation.potential.pppm_h_array[1],
                    simulation.potential.pppm_h_array[2],
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print(
                "Mesh size * Ewald_parameter (h * alpha) = {:2.4f}, {:2.4f}, {:2.4f} ".format(
                    simulation.potential.pppm_h_array[0] * simulation.potential.pppm_alpha_ewald,
                    simulation.potential.pppm_h_array[1] * simulation.potential.pppm_alpha_ewald,
                    simulation.potential.pppm_h_array[2] * simulation.potential.pppm_alpha_ewald,
                )
            )
            print(
                "                                        ~ 1/{}, 1/{}, 1/{}".format(
                    int(1.0 / (simulation.potential.pppm_h_array[0] * simulation.potential.pppm_alpha_ewald)),
                    int(1.0 / (simulation.potential.pppm_h_array[1] * simulation.potential.pppm_alpha_ewald)),
                    int(1.0 / (simulation.potential.pppm_h_array[2] * simulation.potential.pppm_alpha_ewald)),
                )
            )
            print(
                "rcut = {:2.4f} a_ws = {:.6e} ".format(
                    simulation.potential.rc / simulation.parameters.a_ws, simulation.potential.rc
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print(
                "No. of PP cells per dimension = {:2}, {:2}, {:2}".format(
                    int(simulation.parameters.box_lengths[0] / simulation.potential.rc),
                    int(simulation.parameters.box_lengths[1] / simulation.potential.rc),
                    int(simulation.parameters.box_lengths[2] / simulation.potential.rc),
                )
            )
            print(
                "No. of particles in PP loop = {:6}".format(
                    int(
                        simulation.parameters.total_num_ptcls
                        / simulation.parameters.box_volume
                        * (3 * simulation.potential.rc) ** 3
                    )
                )
            )
            print(
                "No. of PP neighbors per particle = {:6}".format(
                    int(
                        simulation.parameters.total_num_ptcls
                        / simulation.parameters.box_volume
                        * 4.0
                        / 3.0
                        * np.pi
                        * (simulation.potential.rc) ** 3.0
                    )
                )
            )
            print("PM Force Error = {:.6e}".format(simulation.parameters.pppm_pm_err))
            print("PP Force Error = {:.6e}".format(simulation.parameters.pppm_pp_err))

        elif simulation.potential.method == "pp":
            print(
                "rcut = {:2.4f} a_ws = {:.6e} ".format(
                    simulation.potential.rc / simulation.parameters.a_ws, simulation.potential.rc
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print(
                "No. of PP cells per dimension = {:2}, {:2}, {:2}".format(
                    int(simulation.parameters.box_lengths[0] / simulation.potential.rc),
                    int(simulation.parameters.box_lengths[1] / simulation.potential.rc),
                    int(simulation.parameters.box_lengths[2] / simulation.potential.rc),
                )
            )
            print(
                "No. of particles in PP loop = {:6}".format(
                    int(
                        simulation.parameters.total_num_ptcls
                        / simulation.parameters.box_volume
                        * (3 * simulation.potential.rc) ** 3
                    )
                )
            )
            print(
                "No. of PP neighbors per particle = {:6}".format(
                    int(
                        simulation.parameters.total_num_ptcls
                        * 4.0
                        / 3.0
                        * np.pi
                        * (simulation.potential.rc / simulation.parameters.box_lengths.min()) ** 3.0
                    )
                )
            )

        print("Tot Force Error = {:.6e}".format(simulation.parameters.force_error))

    @staticmethod
    def potential_info(simulation):
        """
        Print potential information.

        Parameters
        ----------
        simulation: sarkas.core.Simulation
            Simulation's parameters.

        """
        if simulation.potential.type == "yukawa":
            print(
                "electron temperature = {:1.4e} [K] = {:1.4e} [eV]".format(
                    simulation.parameters.electron_temperature,
                    simulation.parameters.electron_temperature / simulation.parameters.eV2K,
                )
            )
            print("kappa = {:.4f}".format(simulation.parameters.a_ws / simulation.parameters.lambda_TF))
            print("Gamma_eff = {:.2f}".format(simulation.parameters.coupling_constant))

        elif simulation.potential.type == "egs":
            # print('electron temperature = {:1.4e} [K] = {:1.4e} eV'.format(
            #     simulation.parameters.electron_temperature,
            #     simulation.parameters.electron_temperature / simulation.parameters.eV2K))
            print("kappa = {:.4f}".format(simulation.parameters.a_ws / simulation.parameters.lambda_TF))
            print("SGA Correction factor: lmbda = {:.4f}".format(simulation.potential.lmbda))
            # print('lambda_TF = {:1.4e} '.format(simulation.parameters.lambda_TF), end='')
            # print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print("nu = {:.4f}".format(simulation.parameters.nu))
            if simulation.parameters.nu < 1:
                print("Exponential decay:")
                print("lambda_p = {:.6e} ".format(simulation.parameters.lambda_p), end="")
                print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
                print("lambda_m = {:.6e} ".format(simulation.parameters.lambda_m), end="")
                print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
                print("alpha = {:.4f}".format(simulation.parameters.alpha))
                # print('Theta = {:1.4e}'.format(simulation.parameters.electron_degeneracy_parameter))
                print("b = {:.4f}".format(simulation.parameters.b))

            else:
                print("Oscillatory potential:")
                print("gamma_p = {:.6e} ".format(simulation.parameters.gamma_p), end="")
                print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
                print("gamma_m = {:.6e} ".format(simulation.parameters.gamma_m), end="")
                print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
                print("alpha = {:.4f}".format(simulation.parameters.alphap))
                print("b = {:.4f}".format(simulation.parameters.b))

            print("Gamma_eff = {:4.2f}".format(simulation.parameters.coupling_constant))

        elif simulation.potential.type == "coulomb":
            print("Effective Coupling constant: Gamma_eff = {:4.2f}".format(simulation.parameters.coupling_constant))
            print("Short-range Cutoff radius: a_rs = {:.6e} ".format(simulation.potential.a_rs), end="")
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            # simulation.parameters.pretty_print()

        elif simulation.potential.type == "lj":
            print("epsilon = {:.6e}".format(simulation.potential.matrix[0, 0, 0]))
            print("sigma = {:.6e}".format(simulation.potential.matrix[1, 0, 0]))
            print("Gamma_eff = {:.2f}".format(simulation.parameters.coupling_constant))

        elif simulation.potential.type == "qsp":
            print("QSP type: {}".format(simulation.potential.qsp_type))
            print("Pauli term: {}".format(simulation.potential.qsp_pauli))
            print(
                "e de Broglie wavelength = {:.4f} a_ws = {:.6e} ".format(
                    np.sqrt(2.0) * np.pi / (simulation.potential.matrix[1, 0, 0] * simulation.parameters.a_ws),
                    np.sqrt(2.0) * np.pi / simulation.potential.matrix[1, 0, 0],
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print(
                "ion de Broglie wavelength  = {:.4f} a_ws = {:.6e} ".format(
                    np.sqrt(2.0) * np.pi / (simulation.potential.matrix[1, 1, 1] * simulation.parameters.a_ws),
                    np.sqrt(2.0) * np.pi / simulation.potential.matrix[1, 1, 1],
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print(
                "e-e screening length = {:.4f} a_ws = {:.6e} ".format(
                    1.0 / (simulation.potential.matrix[1, 0, 0] * simulation.parameters.a_ws),
                    1.0 / simulation.potential.matrix[1, 0, 0],
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print(
                "i-i screening length = {:.4f} a_ws = {:.6e} ".format(
                    1.0 / (simulation.potential.matrix[1, 1, 1] * simulation.parameters.a_ws),
                    1.0 / simulation.potential.matrix[1, 1, 1],
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print(
                "e-i screening length = {:.4f} a_ws = {:.6e} ".format(
                    1.0 / (simulation.potential.matrix[1, 0, 1] * simulation.parameters.a_ws),
                    1.0 / simulation.potential.matrix[1, 0, 1],
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")
            print("e-i coupling constant = {:.4f} ".format(simulation.parameters.coupling_constant))

        elif simulation.potential.type == "hs_yukawa":
            print(
                "electron temperature = {:1.4e} [K] = {:1.4e} [eV]".format(
                    simulation.parameters.electron_temperature,
                    simulation.parameters.electron_temperature / simulation.parameters.eV2K,
                )
            )
            print("kappa = {:.4f}".format(simulation.parameters.a_ws / simulation.parameters.lambda_TF))
            print(
                "short range cutoff = {:.2f} a_ws = {:.4e} ".format(
                    simulation.potential.matrix[-1, 0, 0] / simulation.parameters.a_ws,
                    simulation.potential.matrix[-1, 0, 0],
                ),
                end="",
            )
            print("[cm]" if simulation.parameters.units == "cgs" else "[m]")

            print("Gamma_eff = {:.2f}".format(simulation.parameters.coupling_constant))

    def setup_checkpoint(self, params, species):
        """
        Assign attributes needed for saving dumps.

        Parameters
        ----------
        params: sarkas.core.Parameters
            General simulation parameters.

        species: sarkas.core.Species
            List of Species classes.

        """
        self.dt = params.dt
        self.a_ws = params.a_ws
        self.total_num_ptcls = params.total_num_ptcls
        self.total_plasma_frequency = params.total_plasma_frequency
        self.species_names = np.copy(params.species_names)
        self.coupling = params.coupling_constant * params.T_desired

        # Check whether energy files exist already
        if not os.path.exists(self.prod_energy_filename):
            # Create the Energy file
            dkeys = ["Time", "Total Energy", "Total Kinetic Energy", "Potential Energy", "Temperature"]
            if len(species) > 1:
                for i, sp in enumerate(species):
                    dkeys.append("{} Kinetic Energy".format(sp.name))
                    dkeys.append("{} Potential Energy".format(sp.name))
                    dkeys.append("{} Temperature".format(sp.name))
            data = dict.fromkeys(dkeys)

            with open(self.prod_energy_filename, "w+") as f:
                w = csv.writer(f)
                w.writerow(data.keys())

        if not os.path.exists(self.eq_energy_filename) and not params.load_method[-7:] == "restart":
            # Create the Energy file
            dkeys = ["Time", "Total Energy", "Total Kinetic Energy", "Potential Energy", "Temperature"]
            if len(species) > 1:
                for i, sp_name in enumerate(params.species_names):
                    dkeys.append("{} Kinetic Energy".format(sp_name))
                    dkeys.append("{} Potential Energy".format(sp_name))
                    dkeys.append("{} Temperature".format(sp_name))
            data = dict.fromkeys(dkeys)

            with open(self.eq_energy_filename, "w+") as f:
                w = csv.writer(f)
                w.writerow(data.keys())

        if self.electrostatic_equilibration:
            if not os.path.exists(self.mag_energy_filename) and not params.load_method[-7:] == "restart":
                # Create the Energy file
                dkeys = ["Time", "Total Energy", "Total Kinetic Energy", "Potential Energy", "Temperature"]
                if len(species) > 1:
                    for i, sp_name in enumerate(params.species_names):
                        dkeys.append("{} Kinetic Energy".format(sp_name))
                        dkeys.append("{} Potential Energy".format(sp_name))
                        dkeys.append("{} Temperature".format(sp_name))
                data = dict.fromkeys(dkeys)

                with open(self.mag_energy_filename, "w+") as f:
                    w = csv.writer(f)
                    w.writerow(data.keys())

    def save_pickle(self, simulation):
        """
        Save all simulations parameters in pickle files.
        """
        file_list = ["parameters", "integrator", "thermostat", "potential", "species"]

        # Redirect to the correct process folder
        if self.process == "preprocessing":
            indx = 0
        else:
            # Note that Postprocessing needs the link to simulation's folder
            # because that is where I look for energy files and pickle files
            indx = 1

        for fl in file_list:
            filename = os.path.join(self.processes_dir[indx], fl + ".pickle")
            pickle_file = open(filename, "wb")
            pickle.dump(simulation.__dict__[fl], pickle_file)
            pickle_file.close()

    def read_pickle(self, process):
        """
        Read pickle files containing all the simulation information.

        Parameters
        ----------
        process: cls
            Simulation's parameters. It can be one of three (sarkas.tools.PreProcess,
            sarkas.core.Simulation, sarkas.tools.PostProcess)
        """
        import copy as py_copy

        file_list = ["parameters", "integrator", "thermostat", "potential", "species"]

        # Redirect to the correct process folder
        if self.process == "preprocessing":
            indx = 0
        else:
            # Note that Postprocessing needs the link to simulation's folder
            # because that is where I look for energy files and pickle files
            indx = 1

        for fl in file_list:
            filename = os.path.join(self.processes_dir[indx], fl + ".pickle")
            data = np.load(filename, allow_pickle=True)
            process.__dict__[fl] = py_copy.copy(data)

    def read_pickle_single(self, class_to_read):
        """
        Read the desired pickle file.

        Parameters
        ----------
        class_to_read : str
            Name of the class to read.

        Returns
        -------
        : cls
            Desired class.

        """
        import copy as py_copy

        # Redirect to the correct process folder
        if self.process == "preprocessing":
            indx = 0
        else:
            # Note that Postprocessing needs the link to simulation's folder
            # because that is where I look for energy files and pickle files
            indx = 1

        filename = os.path.join(self.processes_dir[indx], class_to_read + ".pickle")
        data = np.load(filename, allow_pickle=True)
        return py_copy.copy(data)

    def dump(self, phase, ptcls, it):
        """
        Save particles' data to binary file for future restart.

        Parameters
        ----------
        phase: str
            Simulation phase.

        ptcls: sarkas.core.Particles
            Particles data.

        it : int
            Timestep number.
        """
        if phase == "production":
            ptcls_file = self.prod_ptcls_filename + str(it)
            tme = it * self.dt
            np.savez(
                ptcls_file,
                id=ptcls.id,
                names=ptcls.names,
                pos=ptcls.pos,
                vel=ptcls.vel,
                acc=ptcls.acc,
                cntr=ptcls.pbc_cntr,
                rdf_hist=ptcls.rdf_hist,
                virial=ptcls.virial,
                time=tme,
            )

            energy_file = self.prod_energy_filename

        elif phase == "equilibration":
            ptcls_file = self.eq_ptcls_filename + str(it)
            tme = it * self.dt
            np.savez(
                ptcls_file,
                id=ptcls.id,
                names=ptcls.names,
                pos=ptcls.pos,
                vel=ptcls.vel,
                acc=ptcls.acc,
                virial=ptcls.virial,
                time=tme,
            )

            energy_file = self.eq_energy_filename

        elif phase == "magnetization":
            ptcls_file = self.mag_ptcls_filename + str(it)
            tme = it * self.dt
            np.savez(
                ptcls_file,
                id=ptcls.id,
                names=ptcls.names,
                pos=ptcls.pos,
                vel=ptcls.vel,
                acc=ptcls.acc,
                virial=ptcls.virial,
                time=tme,
            )

            energy_file = self.mag_energy_filename

        kinetic_energies, temperatures = ptcls.kinetic_temperature()
        potential_energies = ptcls.potential_energies()
        # Save Energy data
        data = {
            "Time": it * self.dt,
            "Total Energy": np.sum(kinetic_energies) + ptcls.potential_energy,
            "Total Kinetic Energy": np.sum(kinetic_energies),
            "Potential Energy": ptcls.potential_energy,
            "Total Temperature": ptcls.species_num.transpose() @ temperatures / ptcls.total_num_ptcls,
        }
        if len(temperatures) > 1:
            for sp, kin in enumerate(kinetic_energies):
                data["{} Kinetic Energy".format(self.species_names[sp])] = kin
                data["{} Potential Energy".format(self.species_names[sp])] = potential_energies[sp]
                data["{} Temperature".format(self.species_names[sp])] = temperatures[sp]

        with open(energy_file, "a") as f:
            w = csv.writer(f)
            w.writerow(data.values())

    def dump_xyz(self, phase="production"):
        """
        Save the XYZ file by reading Sarkas dumps.

        Parameters
        ----------
        phase : str
            Phase from which to read dumps. 'equilibration' or 'production'.

        dump_skip : int
            Interval of dumps to skip. Default = 1

        """

        if phase == "equilibration":
            self.xyz_filename = os.path.join(self.equilibration_dir, "pva_" + self.job_id + ".xyz")
            dump_dir = self.eq_dump_dir

        else:
            self.xyz_filename = os.path.join(self.production_dir, "pva_" + self.job_id + ".xyz")
            dump_dir = self.prod_dump_dir

        f_xyz = open(self.xyz_filename, "w+")

        if not hasattr(self, "a_ws"):
            params = self.read_pickle_single("parameters")
            self.a_ws = params.a_ws
            self.total_num_ptcls = params.total_num_ptcls
            self.total_plasma_frequency = params.total_plasma_frequency

        # Rescale constants. This is needed since OVITO has a small number limit.
        pscale = 1.0 / self.a_ws
        vscale = 1.0 / (self.a_ws * self.total_plasma_frequency)
        ascale = 1.0 / (self.a_ws * self.total_plasma_frequency ** 2)

        # Read the list of dumps and sort them in the correct (natural) order
        dumps = os.listdir(dump_dir)
        dumps.sort(key=num_sort)
        for dump in tqdm(dumps, disable=not self.verbose):
            data = self.read_npz(dump_dir, dump)
            data["pos_x"] *= pscale
            data["pos_y"] *= pscale
            data["pos_z"] *= pscale

            data["vel_x"] *= vscale
            data["vel_y"] *= vscale
            data["vel_z"] *= vscale

            data["acc_x"] *= ascale
            data["acc_y"] *= ascale
            data["acc_z"] *= ascale

            f_xyz.writelines("{0:d}\n".format(self.total_num_ptcls))
            f_xyz.writelines("name x y z vx vy vz ax ay az\n")
            np.savetxt(f_xyz, data, fmt="%s %.6e %.6e %.6e %.6e %.6e %.6e %.6e %.6e %.6e")

        f_xyz.close()

    @staticmethod
    def read_npz(fldr, it):
        """
        Load particles' data from dumps.

        Parameters
        ----------
        fldr : str
            Folder containing dumps.

        it : str
            Timestep to load.

        Returns
        -------
        struct_array : numpy.ndarray
            Structured data array.

        """

        file_name = os.path.join(fldr, it)
        data = np.load(file_name, allow_pickle=True)
        # Dev Notes: the old way of saving the xyz file by
        # np.savetxt(f_xyz, np.c_[data["names"],data["pos"] ....]
        # , fmt="%10s %.6e %.6e %.6e %.6e %.6e %.6e %.6e %.6e %.6e")
        # was not working, because the columns of np.c_[] all have the same data type <U32
        # which is in conflict with the desired fmt. i.e. data["names"] was not recognized as a string.
        # So I have to create a new structured array and pass this. I could not think of a more Pythonic way.
        struct_array = np.zeros(
            data["names"].size,
            dtype=[
                ("names", "U6"),
                ("pos_x", np.float64),
                ("pos_y", np.float64),
                ("pos_z", np.float64),
                ("vel_x", np.float64),
                ("vel_y", np.float64),
                ("vel_z", np.float64),
                ("acc_x", np.float64),
                ("acc_y", np.float64),
                ("acc_z", np.float64),
            ],
        )
        struct_array["names"] = data["names"]
        struct_array["pos_x"] = data["pos"][:, 0]
        struct_array["pos_y"] = data["pos"][:, 1]
        struct_array["pos_z"] = data["pos"][:, 2]

        struct_array["vel_x"] = data["vel"][:, 0]
        struct_array["vel_y"] = data["vel"][:, 1]
        struct_array["vel_z"] = data["vel"][:, 2]

        struct_array["acc_x"] = data["acc"][:, 0]
        struct_array["acc_y"] = data["acc"][:, 1]
        struct_array["acc_z"] = data["acc"][:, 2]

        return struct_array


def alpha_to_int(text):
    return int(text) if text.isdigit() else text


def num_sort(text):
    """
    Method copied from
    https://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside

    Notes
    -----
    originally from http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)

    Parameters
    ----------
    text : str
        Text to be split into str and int

    Returns
    -------
     : list
        List containing text and integers

    """

    return [alpha_to_int(c) for c in re.split(r"(\d+)", text)]


def convert_bytes(tot_bytes):
    GB, rem = divmod(tot_bytes, 1024 * 1024 * 1024)
    MB, rem = divmod(rem, 1024 * 1024)
    KB, rem = divmod(rem, 1024)

    return [GB, MB, KB, rem]



"""Module of Exceptions and warnings specific to Sarkas."""

# __all__ = [
#     "SarkasError",
#     "ParticlesError",
#     "SarkasWarning",
#     "PhysicsWarning",
#     "AlgorithmWarning",
# ]

# ------------------------------------------------------------------------------
#   Exceptions
# ------------------------------------------------------------------------------


class SarkasError(Exception):
    """
    Base class of Sarkas custom errors.
    All custom exceptions raised by Sarkas should inherit from this
    class and be defined in this module.
    """


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


# ^^^^^^^^^^^^ Base Exceptions should be defined above this comment ^^^^^^^^^^^^


class AlgorithmError(SarkasError):
    """The base error for errors related to the chosen algorithm."""


class ParticlesError(SarkasError):
    """The base error for errors related to the Particles class."""


# ------------------------------------------------------------------------------
#   Warnings
# ------------------------------------------------------------------------------


class SarkasWarning(Warning):
    """
    Base class of Sarkas custom warnings.
    All Sarkas custom warnings should inherit from this class and be
    defined in this module.
    Warnings should be issued using `warnings.warn`, which will not break
    execution if unhandled.
    """

    pass


class PhysicsWarning(Warning):
    """The base warning for warnings related to non-physical situations."""


# ^^^^^^^^^^^^^ Base Warnings should be defined above this comment ^^^^^^^^^^^^^


class AlgorithmWarning(SarkasWarning):
    """The base warning for warnings related to the used algorithm."""
